# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the Apache License, Version 2.0
# found in the LICENSE file in the root directory of this source tree.

import itertools
import math

import torch
import torch.nn as nn
import torch.nn.functional as F

_DINOV2_BASE_URL = "https://dl.fbaipublicfiles.com/dinov2"


def _make_dinov2_model_name(arch_name: str, patch_size: int, num_register_tokens: int = 0) -> str:
    """Generates a DINOv2 model name based on architecture name, patch size, and number of register tokens.

    Args:
        arch_name (str): The name of the architecture.
        patch_size (int): The size of the patches.
        num_register_tokens (int, optional): The number of register tokens. Defaults to 0.

    Returns:
        str: The formatted DINOv2 model name.
    """
    compact_arch_name = arch_name.replace("_", "")[:4]
    registers_suffix = f"_reg{num_register_tokens}" if num_register_tokens else ""
    return f"dinov2_{compact_arch_name}{patch_size}{registers_suffix}"


class CenterPadding(nn.Module):
    """A module that applies center padding to input tensors to ensure dimensions are multiples of a specified value."""

    def __init__(self, multiple: int):
        """Initializes the CenterPadding module.

        Args:
            multiple (int): The value to which the dimensions should be a multiple of.
        """
        super().__init__()
        self.multiple = multiple

    def _get_pad(self, size: int) -> tuple:
        """Calculates the padding required to make the size a multiple of the specified value.

        Args:
            size (int): The original size.

        Returns:
            tuple: A tuple containing the left and right padding sizes.
        """
        new_size = math.ceil(size / self.multiple) * self.multiple
        pad_size = new_size - size
        pad_size_left = pad_size // 2
        pad_size_right = pad_size - pad_size_left
        return pad_size_left, pad_size_right

    @torch.inference_mode()
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Applies center padding to the input tensor.

        Args:
            x (torch.Tensor): The input tensor to be padded.

        Returns:
            torch.Tensor: The padded tensor.
        """
        pads = list(itertools.chain.from_iterable(self._get_pad(m) for m in x.shape[:1:-1]))
        output = F.pad(x, pads)
        return output



# Copyright (c) Facebook, Inc. and its affiliates. (http://www.facebook.com)
# -*- coding: utf-8 -*-
"""mapillary.interface
~~~~~~~~~~~~~~~~~~~

This module implements the basic functionalities of the Mapillary Python SDK, a Python
implementation of the Mapillary API v4. For more information, please check out
https://www.mapillary.com/developer/api-documentation/

- Copyright: (c) 2021 Facebook
- License: MIT LICENSE
"""
import json
import os
from typing import Union

import requests

import zensvi.download.mapillary.controller.detection as detection
import zensvi.download.mapillary.controller.feature as feature
import zensvi.download.mapillary.controller.image as image
import zensvi.download.mapillary.controller.save as save
from zensvi.download.mapillary.models.config import Config
from zensvi.download.mapillary.models.exceptions import InvalidOptionError
from zensvi.download.mapillary.models.geojson import Coordinates, GeoJSON
from zensvi.download.mapillary.utils.auth import auth, set_token


def configure_mapillary_settings(**kwargs):
    """A function allowing the user to configure the Mapillary settings for the session.
    Takes no arguments and sets a global variable used by other functions making API requests.
    For more information what the details of authentication, please check out the blog post at Mapillary.
    https://blog.mapillary.com/update/2021/06/23/getting-started-with-the-new-mapillary-api-v4.html

    Args:
      kwargs: Configuration options
      **kwargs:

    Returns:
      : None
      Usage: :

    >>> import mapillary as mly
        >>> mly.interface.configure_mapillary_settings()
        >>> mly.interface.configure_mapillary_settings(use_strict=True)
    """
    return Config(kwargs)


def set_access_token(token: str):
    """A function allowing the user to set an access token for the session.
    Takes token as an argument and sets a global variable used by other functions making API requests.
    For more information what the details of authentication, please check out the blog post at Mapillary.
    https://blog.mapillary.com/update/2021/06/23/getting-started-with-the-new-mapillary-api-v4.html

    Args:
      token: The access token to set
      token: str:
      token: str:

    Returns:
      : None
      Usage: :

    >>> import mapillary as mly
        >>> mly.interface.set_access_token('CLIENT_TOKEN_HERE')
    """
    return set_token(token)


@auth()
def get_image_close_to(latitude=-122.1504711, longitude=37.485073, **kwargs):
    """Function that takes a longitude, latitude as argument and outputs the near
    images. This makes an API call with the token set in set_access_token and returns a
    JSON object.

    Args:
      longitude(float or double, optional): The longitude (Default value = 37.485073)
      latitude(float or double, optional): The latitude (Default value = -122.1504711)
      kwargs.fields(list): A list of options, either as ['all'], or a
    list of fields. See https://www.mapillary.com/developer/api-
    documentation/, under 'Fields' for more insight.
      kwargs.zoom(int): The zoom level of the tiles to obtain,
    defaults to 14
      kwargs.radius(float or int or double): The radius of the images
    obtained from a center center
      kwargs.image_type(str): The tile image_type to be obtained,
    either as 'flat', 'pano' (panoramic), or 'both'. See
    https://www.mapillary.com/developer/api-documentation/ under
    'image_type Tiles' for more information
      kwargs.min_captured_at(str): The min date. Format from 'YYYY',
    to 'YYYY-MM-DDTHH:MM:SS'
      kwargs.max_captured_at(str): The max date. Format from 'YYYY',
    to 'YYYY-MM-DDTHH:MM:SS'
      kwargs.org_id(int): The organization id, ID of the organization
    this image (or sets of images) belong to. It can be absent.
    Thus, default is -1 (None)
      **kwargs:

    Returns:
      dict: GeoJSON
      Usage: :

    >>> import mapillary as mly
        >>> mly.interface.set_access_token('CLIENT_TOKEN_HERE')
        >>> mly.interface.get_image_close_to(longitude=31, latitude=30)
        ... {'type': 'FeatureCollection', 'features': [{'type': 'Feature',
        'geometry': {'type': 'Point', 'coordinates': [30.9912246465683,
        29.99794091267283]}, 'properties': {'captured_at': 1621008070596,
        'compass_angle': 322.56726074219, 'id': 499412381300321, 'is_pano':
        False, 'sequence_id': '94afmyyhq85xd9bi8p44ve'}} ...
    """
    return image.get_image_close_to_controller(
        latitude=latitude,
        longitude=longitude,
        kwargs=kwargs,
    )


@auth()
def get_image_looking_at(
    at: dict,
    **filters: dict,
) -> GeoJSON:
    """Function that takes two sets of latitude and longitude, where the 2nd set is the
    "looking at" location from 1st set's perspective argument and outputs the near
    images. This makes an API call with the token set in set_access_token and returns a
    JSON object.

    Args:
      at(dict): The coordinate sets to where a certain point is being
    looked at
    Format::
      at: dict:
      **filters: dict:
      at: dict:
      **filters: dict:
      at: dict:
      **filters: dict:

    Returns:
      GeoJSON: The GeoJSON response containing relevant features
      Usage: :

    >>> {
                ...     'lng': 'longitude',
                ...     'lat': 'latitude'
                ... }
        filters.min_captured_at (str): The minimum date to filter till
        filters.max_captured_at (str): The maximum date to filter upto
        filters.radius (float): The radius that the geometry points will
            lie in
        filters.image_type (str): Either 'pano', 'flat' or 'all'
        filters.organization_id (str): The organization to retrieve the
            data for

        >>> import mapillary as mly
        >>> mly.interface.set_access_token('MLY|XXX')
        >>> data = mly.interface.get_image_looking_at(
        ...         at={
        ...             'lng': 12.955075073889,
        ...             'lat': 48.053805939722,
        ...         },
        ...         radius = 5000,
        ...     )
        >>> data
        ... {'type': 'FeatureCollection', 'features': [{'type': 'Feature', 'geometry': {'type':
        ... 'Point', 'coordinates': [12.954479455947876, 48.05091893670834]}, 'properties':
        ... {'captured_at': 1612606959408, 'compass_angle': 21.201110839844, 'id': 1199705400459580,
        ... 'is_pano': False, 'sequence_id': 'qrrqtke4a6vtygyc7w8rzc'}}, ... }
    """
    return image.get_image_looking_at_controller(
        at=at,
        filters=filters,
    )


@auth()
def is_image_being_looked_at(
    at: Union[dict, Coordinates, list],
    **filters: dict,
) -> bool:
    """Function that two sets of coordinates and returns whether the image  with
    coordinates of "at" is looked at or not by the image with coordinates of "looker".

    Args:
      at(Union[dict): The coordinate sets to where a certain point is being looked
    at
    Format::
      at: Union[dict:
      Coordinates:
      list]:
      **filters: dict:
      at: Union[dict:
      **filters: dict:
      at: Union[dict:
      **filters: dict:

    Returns:
      bool: True if the image is looked at, False otherwise
      Usage: :

    >>> at_dict = {
                ...     'lng': 'longitude',
                ...     'lat': 'latitude'
                ... }
                >>> at_list = [12.954940544167, 48.0537894275]
                >>> from mapillary.models.geojson import Coordinates
                >>> at_coord: Coordinates = Coordinates(lng=12.954940544167, lat=48.0537894275)

        >>> import mapillary as mly
        >>> mly.interface.set_access_token('MLY|XXX')
        >>> mly.interface.is_image_being_looked_at(
        ...         at={
        ...             'lng': 12.955075073889,
        ...             'lat': 48.053805939722,
        ...         },
        ...         radius=50,
        ...     )
        ... True
        >>> # OR
        >>> from mapillary.models.geojson import Coordinates
        >>> mly.interface.is_image_looked_at(
        ...         at=Coordinates(lng=11.954940544167, lat=46.0537894275),
        ...         radius=50,
        ...     )
        ... True
    """
    return image.is_image_being_looked_at_controller(at=at, filters=filters)


@auth()
def get_detections_with_image_id(image_id: int, fields: list = []):
    """Extracting all the detections within an image using an image key.

    Args:
      image_id(int): The image key as the argument
      fields(list): The fields possible for the detection endpoint.
    Please see https://www.mapillary.com/developer/api-
    documentation for more information
      image_id: int:
      fields: list:  (Default value = [])
      image_id: int:
      fields: list:  (Default value = [])
      image_id: int:
      fields: list:  (Default value = [])

    Returns:
      dict: The GeoJSON in response
      Usage: :

    >>> import mapillary as mly
        >>> mly.interface.set_access_token('CLIENT_TOKEN_HERE')
        >>> mly.interface.get_detections_with_image_id(image_id=1933525276802129)
        ... {"data":[{"created_at":"2021-05-20T17:49:01+0000","geometry":
        ... "GjUKBm1weS1vchIVEgIAABgDIg0JhiekKBoqAABKKQAPGgR0eXBlIgkKB3BvbHlnb24ogCB4AQ==","image"
        ... :{"geometry":{"type":"Point","coordinates":[-97.743279722222,30.270651388889]},"id":
        ... "1933525276802129"},"value":"regulatory--no-parking--g2","id":"1942105415944115"},
        ... {"created_at":"2021-05-20T18:40:21+0000","geometry":
        ... "GjYKBm1weS1vchIWEgIAABgDIg4J7DjqHxpWAADiAVUADxoEdHlwZSIJCgdwb2x5Z29uKIAgeAE=",
        ... "image":{"geometry":{"type":"Point","coordinates":[-97.743279722222,30.270651388889]},
        ... "id":"1933525276802129"},"value":"information--parking--g1","id":"1942144785940178"},
        ... , ...}
    """
    return detection.get_image_detections_controller(
        image_id=image_id,
        fields=fields,
    )


@auth()
def get_detections_with_map_feature_id(map_feature_id: str, fields: list = None) -> GeoJSON:
    """Extracting all detections made for a map feature key.

    Args:
      map_feature_id(int): A map feature key as the argument
      fields(list): The fields possible for the detection endpoint.
    Please see https://www.mapillary.com/developer/api-
    documentation for more information
      map_feature_id: str:
      fields: list:  (Default value = None)
      map_feature_id: str:
      fields: list:  (Default value = None)
      map_feature_id: str:
      fields: list:  (Default value = None)

    Returns:
      GeoJSON: The GeoJSON in response
      Usage: :

    >>> import mapillary as mly
        >>> mly.interface.set_access_token('MLY|XXX')
        >>> mly.interface.get_detections_with_map_feature_id(map_feature_id='1933525276802129')
        ...     File "/home/saif/MLH/mapillary-python-sdk/mapillary/controller/rules/verify.py",
        ...         line 227, in valid_id
        ...             raise InvalidOptionError(
        ... mly.models.exceptions.InvalidOptionError: InvalidOptionError: Given id value,
        ...     "Id: 1933525276802129, image: False" while possible id options, [Id is image_id
        ...     AND image is True, key is map_feature_id ANDimage is False]
    """
    return detection.get_map_feature_detections_controller(
        map_feature_id=map_feature_id,
        fields=fields,
    )


@auth()
def image_thumbnail(image_id: str, resolution: int = 1024, additional_fields: list = None) -> str:
    """Gets the thumbnails of images from the API.

    Args:
      image_id: Image key as the argument
      resolution: Option for the thumbnail size, with available
    resolutions: 256, 1024, and 2048
      image_id: str:
      resolution: int:  (Default value = 1024)
      image_id: str:
      resolution: int:  (Default value = 1024)
      additional_fields: list:  (Default value = None)
      image_id: str:
      resolution: int:  (Default value = 1024)
      additional_fields: list:  (Default value = None)

    Returns:
      str: A URL for the thumbnail
      Usage: :

    >>> import mapillary as mly
        >>> mly.interface.set_access_token('MLY|XXX')
        >>> mly.interface.image_thumbnail(
        ...     image_id='IMAGE_ID_HERE',
        ...     resolution=1024
        ... )
    """
    return image.get_image_thumbnail_controller(image_id, resolution, additional_fields)


@auth()
def images_in_bbox(bbox: dict, **filters) -> str:
    """Gets a complete list of images with custom filter within a BBox.

    Args:
      bbox(dict): Bounding box coordinates
    Format::
    Example filters::
    - max_captured_at
    - min_captured_at
    - image_type: pano, flat, or all
    - compass_angle
    - sequence_id
    - organization_id
      bbox: dict:
      **filters:
      bbox: dict:
      bbox: dict:

    Returns:
      str: Output is a GeoJSON string that represents all the within a
      str: Output is a GeoJSON string that represents all the within a
      str: Output is a GeoJSON string that represents all the within a
      str: Output is a GeoJSON string that represents all the within a
      bbox after passing given filters
      Usage: :

    >>> {
                ...     'west': 'BOUNDARY_FROM_WEST',
                ...     'south': 'BOUNDARY_FROM_SOUTH',
                ...     'east': 'BOUNDARY_FROM_EAST',
                ...     'north': 'BOUNDARY_FROM_NORTH'
                ... }
        **filters (dict): Different filters that may be applied to the
            output

        >>> import mapillary as mly
        >>> mly.interface.set_access_token('MLY|XXX')
        >>> mly.interface.images_in_bbox(
        ...     bbox={
        ...         'west': 'BOUNDARY_FROM_WEST',
        ...         'south': 'BOUNDARY_FROM_SOUTH',
        ...         'east': 'BOUNDARY_FROM_EAST',
        ...         'north': 'BOUNDARY_FROM_NORTH'
        ...     },
        ...     max_captured_at='YYYY-MM-DD HH:MM:SS',
        ...     min_captured_at='YYYY-MM-DD HH:MM:SS',
        ...     image_type='pano',
        ...     compass_angle=(0, 360),
        ...     sequence_id='SEQUENCE_ID',
        ...     organization_id='ORG_ID'
        ... )
    """
    return image.get_images_in_bbox_controller(bounding_box=bbox, layer="image", zoom=14, filters=filters)


@auth()
def sequences_in_bbox(bbox: dict, **filters) -> str:
    """Gets a complete list of all sequences of images that satisfy given filters within
    a BBox.

    Args:
      bbox(dict): Bounding box coordinates
    Example::
    Example filters::
    - max_captured_at
    - min_captured_at
    - image_type: pano, flat, or all
    - org_id
      bbox: dict:
      **filters:
      bbox: dict:
      bbox: dict:

    Returns:
      str: Output is a GeoJSON string that contains all the filtered
      str: Output is a GeoJSON string that contains all the filtered
      str: Output is a GeoJSON string that contains all the filtered
      str: Output is a GeoJSON string that contains all the filtered
      sequences within a bbox. Sequences would NOT be cut at BBox
      str: Output is a GeoJSON string that contains all the filtered
      str: Output is a GeoJSON string that contains all the filtered
      str: Output is a GeoJSON string that contains all the filtered
      sequences within a bbox. Sequences would NOT be cut at BBox
      str: Output is a GeoJSON string that contains all the filtered
      str: Output is a GeoJSON string that contains all the filtered
      sequences within a bbox. Sequences would NOT be cut at BBox
      str: Output is a GeoJSON string that contains all the filtered
      sequences within a bbox. Sequences would NOT be cut at BBox
      boundary, would select all sequences which are partially or
      str: Output is a GeoJSON string that contains all the filtered
      str: Output is a GeoJSON string that contains all the filtered
      str: Output is a GeoJSON string that contains all the filtered
      sequences within a bbox. Sequences would NOT be cut at BBox
      str: Output is a GeoJSON string that contains all the filtered
      str: Output is a GeoJSON string that contains all the filtered
      sequences within a bbox. Sequences would NOT be cut at BBox
      str: Output is a GeoJSON string that contains all the filtered
      sequences within a bbox. Sequences would NOT be cut at BBox
      boundary, would select all sequences which are partially or
      str: Output is a GeoJSON string that contains all the filtered
      str: Output is a GeoJSON string that contains all the filtered
      sequences within a bbox. Sequences would NOT be cut at BBox
      str: Output is a GeoJSON string that contains all the filtered
      sequences within a bbox. Sequences would NOT be cut at BBox
      boundary, would select all sequences which are partially or
      str: Output is a GeoJSON string that contains all the filtered
      sequences within a bbox. Sequences would NOT be cut at BBox
      boundary, would select all sequences which are partially or
      entirely in BBox
      Usage: :

    >>> _ = {
                ...     'west': 'BOUNDARY_FROM_WEST',
                ...     'south': 'BOUNDARY_FROM_SOUTH',
                ...     'east': 'BOUNDARY_FROM_EAST',
                ...     'north': 'BOUNDARY_FROM_NORTH'
                ... }
        **filters (dict): Different filters that may be applied to the
            output

        >>> import mapillary as mly
        >>> mly.interface.set_access_token('MLY|XXX')
        >>> mly.interface.sequences_in_bbox(
        ...     bbox={
        ...         'west': 'BOUNDARY_FROM_WEST',
        ...         'south': 'BOUNDARY_FROM_SOUTH',
        ...         'east': 'BOUNDARY_FROM_EAST',
        ...         'north': 'BOUNDARY_FROM_NORTH'
        ...     },
        ...     max_captured_at='YYYY-MM-DD HH:MM:SS',
        ...     min_captured_at='YYYY-MM-DD HH:MM:SS',
        ...     image_type='pano',
        ...     org_id='ORG_ID'
        ... )
    """
    return image.get_images_in_bbox_controller(bounding_box=bbox, layer="sequence", zoom=14, filters=filters)


@auth()
def map_feature_points_in_bbox(bbox: dict, filter_values: list = None, **filters: dict) -> str:
    """Extracts map feature points within a bounding box (bbox)

    Args:
      bbox(dict): bbox coordinates as the argument
    Example::
    Example::
    Chronological filters,
    - *existed_at*: checks if a feature existed after a certain date depending on the time
    it was first seen at.
    - *existed_before*: filters out the features that existed after a given date
      bbox: dict:
      filter_values: list:  (Default value = None)
      **filters: dict:
      bbox: dict:
      filter_values: list:  (Default value = None)
      **filters: dict:
      bbox: dict:
      filter_values: list:  (Default value = None)
      **filters: dict:

    Returns:
      dict: GeoJSON Object
      Usage: :

    >>> _ = {
                ...     'west': 'BOUNDARY_FROM_WEST',
                ...     'south': 'BOUNDARY_FROM_SOUTH',
                ...     'east': 'BOUNDARY_FROM_EAST',
                ...     'north': 'BOUNDARY_FROM_NORTH'
                ... }
        filter_values (list): a list of filter values supported by the
            API

                >>> _ = ['object--support--utility-pole', 'object--street-light']
        **filters (dict): kwarg filters to be applied on the resulted
            GeoJSON

        >>> import mapillary as mly
        >>> mly.interface.set_access_token('MLY|XXX')
        >>> mly.interface.map_feature_points_in_bbox(
        ...     bbox={
        ...         'west': 'BOUNDARY_FROM_WEST',
        ...         'south': 'BOUNDARY_FROM_SOUTH',
        ...         'east': 'BOUNDARY_FROM_EAST',
        ...         'north': 'BOUNDARY_FROM_NORTH'
        ...     },
        ...     filter_values=['object--support--utility-pole', 'object--street-light'],
        ...     existed_at='YYYY-MM-DD HH:MM:SS',
        ...     existed_before='YYYY-MM-DD HH:MM:SS'
        ... )
    """
    return feature.get_map_features_in_bbox_controller(
        bbox=bbox, filters=filters, filter_values=filter_values, layer="points"
    )


@auth()
def traffic_signs_in_bbox(bbox: dict, filter_values: list = None, **filters: dict) -> str:
    """Extracts traffic signs within a bounding box (bbox)

    Args:
      bbox(dict): bbox coordinates as the argument
    Example::
    Example::
    Chronological filters,
    - *existed_at*: checks if a feature existed after a certain date depending on the time
    it was first seen at.
    - *existed_before*: filters out the features that existed after a given date
      bbox: dict:
      filter_values: list:  (Default value = None)
      **filters: dict:
      bbox: dict:
      filter_values: list:  (Default value = None)
      **filters: dict:
      bbox: dict:
      filter_values: list:  (Default value = None)
      **filters: dict:

    Returns:
      dict: GeoJSON Object
      Usage: :

    >>> {
                ...     'west': 'BOUNDARY_FROM_WEST',
                ...     'south': 'BOUNDARY_FROM_SOUTH',
                ...     'east': 'BOUNDARY_FROM_EAST',
                ...     'north': 'BOUNDARY_FROM_NORTH'
                ... }
        filter_values (list): a list of filter values supported by the
            API,

                >>> ['regulatory--advisory-maximum-speed-limit--g1', 'regulatory--atvs-permitted--g1']
        **filters (dict): kwarg filters to be applied on the resulted
            GeoJSON

        >>> import mapillary as mly
        >>> mly.interface.set_access_token('MLY|XXX')
        >>> mly.interface.traffic_signs_in_bbox(
        ...    bbox={
        ...         'west': 'BOUNDARY_FROM_WEST',
        ...         'south': 'BOUNDARY_FROM_SOUTH',
        ...         'east': 'BOUNDARY_FROM_EAST',
        ...         'north': 'BOUNDARY_FROM_NORTH'
        ...    },
        ...    filter_values=[
        ...        'regulatory--advisory-maximum-speed-limit--g1',
        ...        'regulatory--atvs-permitted--g1'
        ...    ],
        ...    existed_at='YYYY-MM-DD HH:MM:SS',
        ...    existed_before='YYYY-MM-DD HH:MM:SS'
        ... )
    """
    return feature.get_map_features_in_bbox_controller(
        bbox=bbox, filters=filters, filter_values=filter_values, layer="traffic_signs"
    )


@auth()
def images_in_geojson(
    geojson: dict,
    dir_cache: str = None,
    max_workers: int = 1,
    logger=None,
    **filters: dict,
):
    """Extracts all images within a shape.

    Args:
      geojson(dict): A geojson as the shape acting as the query
    extent
      **filters(dict (kwargs): Different filters that may be applied
    to the output, defaults to {}
      filters.max_captured_at(str): The max date. Format from 'YYYY',
    to 'YYYY-MM-DDTHH:MM:SS'
      filters.min_captured_at(str): The min date. Format from 'YYYY',
    to 'YYYY-MM-DDTHH:MM:SS'
      filters.image_type(str): The tile image_type to be obtained,
    either as 'flat', 'pano' (panoramic), or 'all'. See
    https://www.mapillary.com/developer/api-documentation/ under
    'image_type Tiles' for more information
      filters.compass_angle(int): The compass angle of the image
      filters.sequence_id(str): ID of the sequence this image belongs
    to
      filters.organization_id(str): ID of the organization this image
    belongs to. It can be absent
      geojson: dict:
      dir_cache: str:  (Default value = None)
      max_workers: int:  (Default value = 1)
      logger: (Default value = None)
      **filters: dict:
      geojson: dict:
      dir_cache: str:  (Default value = None)
      max_workers: int:  (Default value = 1)
      **filters: dict:
      geojson: dict:
      dir_cache: str:  (Default value = None)
      max_workers: int:  (Default value = 1)
      **filters: dict:

    Returns:
      mapillary.models.geojson.GeoJSON: A GeoJSON object
      Usage: :

    >>> import mapillary as mly
        >>> from mapillary.models.geojson import GeoJSON
        >>> import json
        >>> mly.interface.set_access_token('MLY|YYY')
        >>> data = mly.interface.images_in_geojson(json.load(open('my_geojson.geojson', mode='r')))
        >>> open('output_geojson.geojson', mode='w').write(data.encode())
    """
    return image.geojson_features_controller(
        geojson=geojson,
        is_image=True,
        filters=filters,
        dir_cache=dir_cache,
        max_workers=max_workers,
        logger=logger,
    )


@auth()
def images_in_shape(shape, **filters: dict):
    """Extracts all images within a shape or polygon.

    Format::

    Args:
      shape(dict): A shape that describes features, formatted as a
    geojson
      **filters(dict (kwargs): Different filters that may be applied
    to the output, defaults to {}
      filters.max_captured_at(str): The max date. Format from 'YYYY',
    to 'YYYY-MM-DDTHH:MM:SS'
      filters.min_captured_at(str): The min date. Format from 'YYYY',
    to 'YYYY-MM-DDTHH:MM:SS'
      filters.image_type(str): The tile image_type to be obtained,
    either as 'flat', 'pano' (panoramic), or 'all'. See
    https://www.mapillary.com/developer/api-documentation/ under
    'image_type Tiles' for more information
      filters.compass_angle(int): The compass angle of the image
      filters.sequence_id(str): ID of the sequence this image belongs
    to
      filters.organization_id(str): ID of the organization this image
    belongs to. It can be absent
      **filters: dict:
      **filters: dict:
      **filters: dict:

    Returns:
      mapillary.models.geojson.GeoJSON: A GeoJSON object
      Usage: :

    >>> {
        ...    "type": "FeatureCollection",
        ...     "features": [
        ...        {
        ...             "type": "Feature",
        ...             "properties": {},
        ...             "geometry": {
        ...                 "type": "Polygon",
        ...                 "coordinates": [
        ...                     [
        ...                         [
        ...                             7.2564697265625,
        ...                             43.69716905314008
        ...                         ],
        ...                         ...
        ...                     ]
        ...                 ]
        ...             }
        ...         }
        ...     ]
        ... }

        >>> import mapillary as mly
        >>> import json
        >>> mly.interface.set_access_token('MLY|XXX')
        >>> data = mly.interface.images_in_shape(json.load(open('polygon.geojson', mode='r')))
        >>> open('output_geojson.geojson', mode='w').write(data.encode())
    """
    return image.shape_features_controller(shape=shape, is_image=True, filters=filters)


@auth()
def map_features_in_geojson(geojson: dict, **filters: dict):
    """Extracts all map features within a geojson's boundaries.

    Args:
      geojson(dict): A geojson as the shape acting as the query
    extent
      **filters(dict (kwargs): Different filters that may be applied
    to the output, defaults to {}
      filters.zoom(int): The zoom level of the tiles to obtain,
    defaults to 14
      filters.max_captured_at(str): The max date. Format from 'YYYY',
    to 'YYYY-MM-DDTHH:MM:SS'
      filters.min_captured_at(str): The min date. Format from 'YYYY',
    to 'YYYY-MM-DDTHH:MM:SS'
      filters.image_type(str): The tile image_type to be obtained,
    either as 'flat', 'pano' (panoramic), or 'all'. See
    https://www.mapillary.com/developer/api-documentation/ under
    'image_type Tiles' for more information
      filters.compass_angle(int): The compass angle of the image
      filters.sequence_id(str): ID of the sequence this image belongs
    to
      filters.organization_id(str): ID of the organization this image
    belongs to. It can be absent
      geojson: dict:
      **filters: dict:
      geojson: dict:
      **filters: dict:
      geojson: dict:
      **filters: dict:

    Returns:
      mapillary.models.geojson.GeoJSON: A GeoJSON object
      Usage: :

    >>> import mapillary as mly
        >>> import json
        >>> mly.interface.set_access_token('MLY|YYY')
        >>> data = mly.interface.map_features_in_geojson(
        ...     json.load(
        ...         open('my_geojson.geojson', mode='r')
        ...     )
        ... )
        >>> open('output_geojson.geojson', mode='w').write(data.encode())
    """
    if isinstance(geojson, str):
        if "http" in geojson:
            geojson = json.loads(requests.get(geojson).content.decode("utf-8"))

    return image.geojson_features_controller(
        geojson=geojson,
        is_image=False,
        filters=filters,
    )


@auth()
def map_features_in_shape(shape: dict, **filters: dict):
    """Extracts all map features within a shape/polygon.

    Format::

    Args:
      shape(dict): A shape that describes features, formatted as a
    geojson
      **filters(dict (kwargs): Different filters that may be applied
    to the output, defaults to {}
      filters.zoom(int): The zoom level of the tiles to obtain,
    defaults to 14
      filters.max_captured_at(str): The max date. Format from 'YYYY',
    to 'YYYY-MM-DDTHH:MM:SS'
      filters.min_captured_at(str): The min date. Format from 'YYYY',
    to 'YYYY-MM-DDTHH:MM:SS'
      filters.image_type(str): The tile image_type to be obtained,
    either as 'flat', 'pano' (panoramic), or 'all'. See
    https://www.mapillary.com/developer/api-documentation/ under
    'image_type Tiles' for more information
      filters.compass_angle(int): The compass angle of the image
      filters.sequence_id(str): ID of the sequence this image belongs
    to
      filters.organization_id(str): ID of the organization this image
    belongs to. It can be absent
      shape: dict:
      **filters: dict:
      shape: dict:
      **filters: dict:
      shape: dict:
      **filters: dict:

    Returns:
      mapillary.models.geojson.GeoJSON: A GeoJSON object
      Usage: :

    >>> _ = {
        ...     "type": "FeatureCollection",
        ...     "features": [
        ...         {
        ...             "type": "Feature",
        ...             "properties": {},
        ...             "geometry": {
        ...                 "type": "Polygon",
        ...                 "coordinates": [
        ...                     [
        ...                         [
        ...                             7.2564697265625,
        ...                             43.69716905314008
        ...                         ],
        ...                         ...
        ...                     ]
        ...                 ]
        ...             }
        ...         }
        ...     ]
        ... }

        >>> import mapillary as mly
        >>> import json
        >>> mly.interface.set_access_token('MLY|XXX')
        >>> data = mly.interface.map_features_in_shape(json.load(open('polygon.geojson', mode='r')))
        >>> open('output_geojson.geojson', mode='w').write(data.encode())
    """
    if isinstance(shape, str):
        if "http" in shape:
            shape = json.loads(requests.get(shape).content.decode("utf-8"))

    return image.shape_features_controller(shape=shape, is_image=False, filters=filters)


@auth()
def feature_from_key(key: str, fields: list = []) -> str:
    """Gets a map feature for the given key argument.

    Args:
      key(int): The map feature ID to which will be used to get the
    feature
      fields(list): The fields to include. The field 'geometry' will
    always be included so you do not need to specify it, or if
    you leave it off, it will still be returned.
    Fields::
    1. first_seen_at - timestamp, timestamp of the least recent
    detection contributing to this feature
    2. last_seen_at - timestamp, timestamp of the most recent
    detection contributing to this feature
    3. object_value - string, what kind of map feature it is
    4. object_type - string, either a traffic_sign or point
    5. geometry - GeoJSON Point geometry
    6. images - list of IDs, which images this map feature was derived
    from
    Refer to https://www.mapillary.com/developer/api-documentation/#map-feature for more details
      key: str:
      fields: list:  (Default value = [])
      key: str:
      fields: list:  (Default value = [])
      key: str:
      fields: list:  (Default value = [])

    Returns:
      str: A GeoJSON string that represents the queried feature
      Usage: :

    >>> import mapillary as mly
        >>> mly.interface.set_access_token('MLY|XXX')
        >>> mly.interface.feature_from_key(
        ...     key='VALID_MAP_FEATURE_KEY',
        ...     fields=['object_value']
        ... )
    """
    return feature.get_feature_from_key_controller(key=int(key), fields=fields)


@auth()
def image_from_key(key: str, fields: list = None) -> str:
    """Gets an image for the given key argument.

    Args:
      key(int): The image unique key which will be used for image
    retrieval
      fields(list): The fields to include. The field 'geometry' will
    always be included so you do not need to specify it, or if
    you leave it off, it will still be returned.
    Fields,
    1. altitude - float, original altitude from Exif
    2. atomic_scale - float, scale of the SfM reconstruction around the image
    3. camera_parameters - array of float, intrinsic camera parameters
    4. camera_type - enum, type of camera projection (perspective, fisheye, or
    spherical)
    5. captured_at - timestamp, capture time
    6. compass_angle - float, original compass angle of the image
    7. computed_altitude - float, altitude after running image processing
    8. computed_compass_angle - float, compass angle after running image processing
    9. computed_geometry - GeoJSON Point, location after running image processing
    10. computed_rotation - enum, corrected orientation of the image
    11. exif_orientation - enum, orientation of the camera as given by the exif tag
    (see: https://sylvana.net/jpegcrop/exif_orientation.html)
    12. geometry - GeoJSON Point geometry
    13. height - int, height of the original image uploaded
    14. thumb_256_url - string, URL to the 256px wide thumbnail
    15. thumb_1024_url - string, URL to the 1024px wide thumbnail
    16. thumb_2048_url - string, URL to the 2048px wide thumbnail
    17. merge_cc - int, id of the connected component of images that were aligned
    together
    18. mesh - { id: string, url: string } - URL to the mesh
    19. quality_score - float, how good the image is (experimental)
    20. sequence - string, ID of the sequence
    21. sfm_cluster - { id: string, url: string } - URL to the point cloud
    22. width - int, width of the original image uploaded
    Refer to https://www.mapillary.com/developer/api-documentation/#image for more details
      key: str:
      fields: list:  (Default value = None)
      key: str:
      fields: list:  (Default value = None)
      key: str:
      fields: list:  (Default value = None)

    Returns:
      str: A GeoJSON string that represents the queried image
      Usage: :

    >>> import mapillary as mly
        >>> mly.interface.set_access_token('MLY|XXX')
        >>> mly.interface.image_from_key(
        ...     key='VALID_IMAGE_KEY',
        ...     fields=['captured_at', 'sfm_cluster', 'width']
        ... )
    """
    return image.get_image_from_key_controller(key=int(key), fields=fields)


@auth()
def save_locally(
    geojson_data: str,
    file_path: str = os.path.dirname(os.path.realpath(__file__)),
    file_name: str = None,
    extension: str = "geojson",
) -> None:
    """This function saves the geojson data locally as a file with the given file name,
    path, and format.

    Args:
      geojson_data(str): The GeoJSON data to be stored
      file_path(str): The path to save the data to. Defaults to the
    current directory path
      file_name(str): The name of the file to be saved. Defaults to
    'geojson'
      extension(str): The format to save the data as. Defaults to
    'geojson'
    Note::
      extension(str): The format to save the data as. Defaults to
    'geojson'
    Note::
    Allowed file format values at the moment are,
    - geojson
    - CSV
    *TODO*: More file format will be supported further in developemtn
    *TODO*: Suggestions and help needed at mapillary/mapillary-python-sdk!
      geojson_data: str:
      file_path: str:  (Default value = os.path.dirname(os.path.realpath(__file__)))
      file_name: str:  (Default value = None)
      extension: str:  (Default value = "geojson")
      geojson_data: str:
      file_path: str:  (Default value = os.path.dirname(os.path.realpath(__file__)))
      file_name: str:  (Default value = None)
      extension: str:  (Default value = "geojson")
      geojson_data: str:
      file_path: str:  (Default value = os.path.dirname(os.path.realpath(__file__)))
      file_name: str:  (Default value = None)
      extension: str:  (Default value = "geojson")

    Returns:
      None: None
      Usage: :

    >>> import mapillary as mly
        >>> mly.interface.set_access_token('MLY|XXX')
        >>> mly.interface.save_locally(
        ...     geojson_data=geojson_data,
        ...     file_path=os.path.dirname(os.path.realpath(__file__)),
        ...     file_name='test_geojson',
        ...     extension='geojson'
        ... )
        >>> mly.interface.save_locally(
        ...     geojson_data=geojson_data,
        ...     file_path=os.path.dirname(os.path.realpath(__file__)),
        ...     file_name='local_geometries',
        ...     extension='csv'
        ... )
    """
    # Check if a valid file format was provided
    if extension.lower() not in ["geojson", "csv"]:
        # If not, raise an error
        raise InvalidOptionError(
            param="format",
            value=extension,
            options=["geojson", "csv"],
        )

    return (
        save.save_as_geojson_controller(data=geojson_data, path=file_path, file_name=file_name)
        if extension.lower() == "geojson"
        else save.save_as_csv_controller(data=geojson_data, path=file_path, file_name=file_name)
    )



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the Apache License, Version 2.0
# found in the LICENSE file in the root directory of this source tree.

# References:
#   https://github.com/facebookresearch/dino/blob/main/vision_transformer.py
#   https://github.com/rwightman/pytorch-image-models/tree/master/timm/models/vision_transformer.py

import logging
import math
from functools import partial
from typing import Callable, Sequence, Tuple, Union

import torch
import torch.nn as nn
import torch.utils.checkpoint
from dinov2.layers import MemEffAttention, Mlp
from dinov2.layers import NestedTensorBlock as Block
from dinov2.layers import PatchEmbed, SwiGLUFFNFused
from torch.nn.init import trunc_normal_

logger = logging.getLogger("dinov2")


def named_apply(fn: Callable, module: nn.Module, name="", depth_first=True, include_root=False) -> nn.Module:
    """Applies a function to a module and its children.

    Args:
        fn (Callable): The function to apply.
        module (nn.Module): The module to apply the function to.
        name (str, optional): The name of the module. Defaults to "".
        depth_first (bool, optional): If True, apply function to children before the module itself. Defaults to True.
        include_root (bool, optional): If True, include the root module in the application. Defaults to False.

    Returns:
        nn.Module: The module after applying the function.
    """
    if not depth_first and include_root:
        fn(module=module, name=name)
    for child_name, child_module in module.named_children():
        child_name = ".".join((name, child_name)) if name else child_name
        named_apply(
            fn=fn,
            module=child_module,
            name=child_name,
            depth_first=depth_first,
            include_root=True,
        )
    if depth_first and include_root:
        fn(module=module, name=name)
    return module


class BlockChunk(nn.ModuleList):
    """A chunk of transformer blocks."""

    def forward(self, x):
        """Forward pass through the chunk of blocks.

        Args:
            x (torch.Tensor): Input tensor.

        Returns:
            torch.Tensor: Output tensor after passing through the blocks.
        """
        for b in self:
            x = b(x)
        return x


class DinoVisionTransformer(nn.Module):
    """Vision Transformer model for DINO."""

    def __init__(
        self,
        img_size=224,
        patch_size=16,
        in_chans=3,
        embed_dim=768,
        depth=12,
        num_heads=12,
        mlp_ratio=4.0,
        qkv_bias=True,
        ffn_bias=True,
        proj_bias=True,
        drop_path_rate=0.0,
        drop_path_uniform=False,
        init_values=None,  # for layerscale: None or 0 => no layerscale
        embed_layer=PatchEmbed,
        act_layer=nn.GELU,
        block_fn=Block,
        ffn_layer="mlp",
        block_chunks=1,
        num_register_tokens=0,
        interpolate_antialias=False,
        interpolate_offset=0.1,
    ):
        """Initializes the DinoVisionTransformer.

        Args:
            img_size (int, optional): Input image size. Defaults to 224.
            patch_size (int, optional): Patch size. Defaults to 16.
            in_chans (int, optional): Number of input channels. Defaults to 3.
            embed_dim (int, optional): Embedding dimension. Defaults to 768.
            depth (int, optional): Depth of transformer. Defaults to 12.
            num_heads (int, optional): Number of attention heads. Defaults to 12.
            mlp_ratio (float, optional): Ratio of MLP hidden dimension to embedding dimension. Defaults to 4.0.
            qkv_bias (bool, optional): Enable bias for QKV if True. Defaults to True.
            ffn_bias (bool, optional): Enable bias for FFN if True. Defaults to True.
            proj_bias (bool, optional): Enable bias for projection in attention if True. Defaults to True.
            drop_path_rate (float, optional): Stochastic depth rate. Defaults to 0.0.
            drop_path_uniform (bool, optional): Apply uniform drop rate across blocks. Defaults to False.
            init_values (float, optional): Layer-scale initialization values. Defaults to None.
            embed_layer (nn.Module, optional): Patch embedding layer. Defaults to PatchEmbed.
            act_layer (nn.Module, optional): MLP activation layer. Defaults to nn.GELU.
            block_fn (nn.Module, optional): Transformer block class. Defaults to Block.
            ffn_layer (str, optional): Type of FFN layer. Options: "mlp", "swiglu", "swiglufused", or "identity". Defaults to "mlp".
            block_chunks (int, optional): Split block sequence into block_chunks units for FSDP wrap. Defaults to 1.
            num_register_tokens (int, optional): Number of extra CLS tokens (so-called "registers"). Defaults to 0.
            interpolate_antialias (bool, optional): Flag to apply anti-aliasing when interpolating positional embeddings. Defaults to False.
            interpolate_offset (float, optional): Work-around offset to apply when interpolating positional embeddings. Defaults to 0.1.
        """
        super().__init__()
        norm_layer = partial(nn.LayerNorm, eps=1e-6)

        self.num_features = self.embed_dim = embed_dim  # num_features for consistency with other models
        self.num_tokens = 1
        self.n_blocks = depth
        self.num_heads = num_heads
        self.patch_size = patch_size
        self.num_register_tokens = num_register_tokens
        self.interpolate_antialias = interpolate_antialias
        self.interpolate_offset = interpolate_offset

        self.patch_embed = embed_layer(
            img_size=img_size,
            patch_size=patch_size,
            in_chans=in_chans,
            embed_dim=embed_dim,
        )
        num_patches = self.patch_embed.num_patches

        self.cls_token = nn.Parameter(torch.zeros(1, 1, embed_dim))
        self.pos_embed = nn.Parameter(torch.zeros(1, num_patches + self.num_tokens, embed_dim))
        assert num_register_tokens >= 0
        self.register_tokens = (
            nn.Parameter(torch.zeros(1, num_register_tokens, embed_dim)) if num_register_tokens else None
        )

        if drop_path_uniform is True:
            dpr = [drop_path_rate] * depth
        else:
            dpr = [x.item() for x in torch.linspace(0, drop_path_rate, depth)]  # stochastic depth decay rule

        if ffn_layer == "mlp":
            logger.info("using MLP layer as FFN")
            ffn_layer = Mlp
        elif ffn_layer == "swiglufused" or ffn_layer == "swiglu":
            logger.info("using SwiGLU layer as FFN")
            ffn_layer = SwiGLUFFNFused
        elif ffn_layer == "identity":
            logger.info("using Identity layer as FFN")

            def f(*args, **kwargs):
                """Identity function for FFN layer.

                Args:
                    *args: Variable length argument list.
                    **kwargs: Arbitrary keyword arguments.

                Returns:
                    nn.Identity: Identity layer.
                """
                return nn.Identity()

            ffn_layer = f
        else:
            raise NotImplementedError

        blocks_list = [
            block_fn(
                dim=embed_dim,
                num_heads=num_heads,
                mlp_ratio=mlp_ratio,
                qkv_bias=qkv_bias,
                proj_bias=proj_bias,
                ffn_bias=ffn_bias,
                drop_path=dpr[i],
                norm_layer=norm_layer,
                act_layer=act_layer,
                ffn_layer=ffn_layer,
                init_values=init_values,
            )
            for i in range(depth)
        ]
        if block_chunks > 0:
            self.chunked_blocks = True
            chunked_blocks = []
            chunksize = depth // block_chunks
            for i in range(0, depth, chunksize):
                # this is to keep the block index consistent if we chunk the block list
                chunked_blocks.append([nn.Identity()] * i + blocks_list[i : i + chunksize])
            self.blocks = nn.ModuleList([BlockChunk(p) for p in chunked_blocks])
        else:
            self.chunked_blocks = False
            self.blocks = nn.ModuleList(blocks_list)

        self.norm = norm_layer(embed_dim)
        self.head = nn.Identity()

        self.mask_token = nn.Parameter(torch.zeros(1, embed_dim))

        self.init_weights()

    def init_weights(self):
        """Initializes the weights of the model."""
        trunc_normal_(self.pos_embed, std=0.02)
        nn.init.normal_(self.cls_token, std=1e-6)
        if self.register_tokens is not None:
            nn.init.normal_(self.register_tokens, std=1e-6)
        named_apply(init_weights_vit_timm, self)

    def interpolate_pos_encoding(self, x, w, h):
        """Interpolates the positional encoding based on input dimensions.

        Args:
            x (torch.Tensor): Input tensor.
            w (int): Width of the input.
            h (int): Height of the input.

        Returns:
            torch.Tensor: Interpolated positional encoding.
        """
        previous_dtype = x.dtype
        npatch = x.shape[1] - 1
        N = self.pos_embed.shape[1] - 1
        if npatch == N and w == h:
            return self.pos_embed
        pos_embed = self.pos_embed.float()
        class_pos_embed = pos_embed[:, 0]
        patch_pos_embed = pos_embed[:, 1:]
        dim = x.shape[-1]
        w0 = w // self.patch_size
        h0 = h // self.patch_size
        # we add a small number to avoid floating point error in the interpolation
        # see discussion at https://github.com/facebookresearch/dino/issues/8
        # DINOv2 with register modify the interpolate_offset from 0.1 to 0.0
        w0, h0 = w0 + self.interpolate_offset, h0 + self.interpolate_offset

        sqrt_N = math.sqrt(N)
        sx, sy = float(w0) / sqrt_N, float(h0) / sqrt_N
        patch_pos_embed = nn.functional.interpolate(
            patch_pos_embed.reshape(1, int(sqrt_N), int(sqrt_N), dim).permute(0, 3, 1, 2),
            scale_factor=(sx, sy),
            mode="bicubic",
            antialias=self.interpolate_antialias,
        )

        assert int(w0) == patch_pos_embed.shape[-2]
        assert int(h0) == patch_pos_embed.shape[-1]
        patch_pos_embed = patch_pos_embed.permute(0, 2, 3, 1).view(1, -1, dim)
        return torch.cat((class_pos_embed.unsqueeze(0), patch_pos_embed), dim=1).to(previous_dtype)

    def prepare_tokens_with_masks(self, x, masks=None):
        """Prepares input tokens with optional masks.

        Args:
            x (torch.Tensor): Input tensor.
            masks (torch.Tensor, optional): Masks to apply. Defaults to None.

        Returns:
            torch.Tensor: Prepared tokens.
        """
        B, nc, w, h = x.shape
        x = self.patch_embed(x)
        if masks is not None:
            x = torch.where(masks.unsqueeze(-1), self.mask_token.to(x.dtype).unsqueeze(0), x)

        x = torch.cat((self.cls_token.expand(x.shape[0], -1, -1), x), dim=1)
        x = x + self.interpolate_pos_encoding(x, w, h)

        if self.register_tokens is not None:
            x = torch.cat(
                (
                    x[:, :1],
                    self.register_tokens.expand(x.shape[0], -1, -1),
                    x[:, 1:],
                ),
                dim=1,
            )

        return x

    def forward_features_list(self, x_list, masks_list):
        """Processes a list of inputs through the transformer blocks.

        Args:
            x_list (list): List of input tensors.
            masks_list (list): List of masks corresponding to the inputs.

        Returns:
            list: List of outputs from the transformer blocks.
        """
        x = [self.prepare_tokens_with_masks(x, masks) for x, masks in zip(x_list, masks_list)]
        for blk in self.blocks:
            x = blk(x)

        all_x = x
        output = []
        for x, masks in zip(all_x, masks_list):
            x_norm = self.norm(x)
            output.append(
                {
                    "x_norm_clstoken": x_norm[:, 0],
                    "x_norm_regtokens": x_norm[:, 1 : self.num_register_tokens + 1],
                    "x_norm_patchtokens": x_norm[:, self.num_register_tokens + 1 :],
                    "x_prenorm": x,
                    "masks": masks,
                }
            )
        return output

    def forward_features(self, x, masks=None):
        """Processes the input through the transformer blocks.

        Args:
            x (torch.Tensor): Input tensor.
            masks (torch.Tensor, optional): Masks to apply. Defaults to None.

        Returns:
            dict: Output containing normalized class token, register tokens, patch tokens, prenormalized output, and masks.
        """
        if isinstance(x, list):
            return self.forward_features_list(x, masks)

        x = self.prepare_tokens_with_masks(x, masks)

        for blk in self.blocks:
            x = blk(x)

        x_norm = self.norm(x)
        return {
            "x_norm_clstoken": x_norm[:, 0],
            "x_norm_regtokens": x_norm[:, 1 : self.num_register_tokens + 1],
            "x_norm_patchtokens": x_norm[:, self.num_register_tokens + 1 :],
            "x_prenorm": x,
            "masks": masks,
        }

    def _get_intermediate_layers_not_chunked(self, x, n=1):
        """Retrieves intermediate layers from the transformer blocks.

        Args:
            x (torch.Tensor): Input tensor.
            n (int, optional): Number of last blocks to take. Defaults to 1.

        Returns:
            list: List of intermediate outputs.
        """
        x = self.prepare_tokens_with_masks(x)
        output, total_block_len = [], len(self.blocks)
        blocks_to_take = range(total_block_len - n, total_block_len) if isinstance(n, int) else n
        for i, blk in enumerate(self.blocks):
            x = blk(x)
            if i in blocks_to_take:
                output.append(x)
        assert len(output) == len(blocks_to_take), f"only {len(output)} / {len(blocks_to_take)} blocks found"
        return output

    def _get_intermediate_layers_chunked(self, x, n=1):
        """Retrieves intermediate layers from chunked transformer blocks.

        Args:
            x (torch.Tensor): Input tensor.
            n (int, optional): Number of last blocks to take. Defaults to 1.

        Returns:
            list: List of intermediate outputs.
        """
        x = self.prepare_tokens_with_masks(x)
        output, i, total_block_len = [], 0, len(self.blocks[-1])
        blocks_to_take = range(total_block_len - n, total_block_len) if isinstance(n, int) else n
        for block_chunk in self.blocks:
            for blk in block_chunk[i:]:  # Passing the nn.Identity()
                x = blk(x)
                if i in blocks_to_take:
                    output.append(x)
                i += 1
        assert len(output) == len(blocks_to_take), f"only {len(output)} / {len(blocks_to_take)} blocks found"
        return output

    def get_intermediate_layers(
        self,
        x: torch.Tensor,
        n: Union[int, Sequence] = 1,  # Layers or n last layers to take
        reshape: bool = False,
        return_class_token: bool = False,
        norm=True,
    ) -> Tuple[Union[torch.Tensor, Tuple[torch.Tensor]]]:
        """Gets intermediate layers from the transformer.

        Args:
            x (torch.Tensor): Input tensor.
            n (Union[int, Sequence], optional): Number of layers or specific layers to take. Defaults to 1.
            reshape (bool, optional): If True, reshape the output. Defaults to False.
            return_class_token (bool, optional): If True, return the class token along with outputs. Defaults to False.
            norm (bool, optional): If True, apply normalization to outputs. Defaults to True.

        Returns:
            Tuple[Union[torch.Tensor, Tuple[torch.Tensor]]]: Outputs from the transformer.
        """
        if self.chunked_blocks:
            outputs = self._get_intermediate_layers_chunked(x, n)
        else:
            outputs = self._get_intermediate_layers_not_chunked(x, n)
        if norm:
            outputs = [self.norm(out) for out in outputs]
        class_tokens = [out[:, 0] for out in outputs]
        outputs = [out[:, 1 + self.num_register_tokens :] for out in outputs]
        if reshape:
            B, _, w, h = x.shape
            outputs = [
                out.reshape(B, w // self.patch_size, h // self.patch_size, -1).permute(0, 3, 1, 2).contiguous()
                for out in outputs
            ]
        if return_class_token:
            return tuple(zip(outputs, class_tokens))
        return tuple(outputs)

    def forward(self, *args, is_training=False, **kwargs):
        """Forward pass through the model.

        Args:
            *args: Variable length argument list.
            is_training (bool, optional): If True, return training output. Defaults to False.
            **kwargs: Arbitrary keyword arguments.

        Returns:
            dict or torch.Tensor: Output from the model.
        """
        ret = self.forward_features(*args, **kwargs)
        if is_training:
            return ret
        else:
            return self.head(ret["x_norm_clstoken"])


def init_weights_vit_timm(module: nn.Module, name: str = ""):
    """Initializes weights for ViT model.

    Args:
        module (nn.Module): The module to initialize.
        name (str, optional): Name of the module. Defaults to "".
    """
    if isinstance(module, nn.Linear):
        trunc_normal_(module.weight, std=0.02)
        if module.bias is not None:
            nn.init.zeros_(module.bias)


def vit_small(patch_size=16, num_register_tokens=0, **kwargs):
    """Creates a small Vision Transformer model.

    Args:
        patch_size (int, optional): Patch size. Defaults to 16.
        num_register_tokens (int, optional): Number of extra CLS tokens. Defaults to 0.
        **kwargs: Additional keyword arguments.

    Returns:
        DinoVisionTransformer: The constructed model.
    """
    model = DinoVisionTransformer(
        patch_size=patch_size,
        embed_dim=384,
        depth=12,
        num_heads=6,
        mlp_ratio=4,
        block_fn=partial(Block, attn_class=MemEffAttention),
        num_register_tokens=num_register_tokens,
        **kwargs,
    )
    return model


def vit_base(patch_size=16, num_register_tokens=0, **kwargs):
    """Creates a base Vision Transformer model.

    Args:
        patch_size (int, optional): Patch size. Defaults to 16.
        num_register_tokens (int, optional): Number of extra CLS tokens. Defaults to 0.
        **kwargs: Additional keyword arguments.

    Returns:
        DinoVisionTransformer: The constructed model.
    """
    model = DinoVisionTransformer(
        patch_size=patch_size,
        embed_dim=768,
        depth=12,
        num_heads=12,
        mlp_ratio=4,
        block_fn=partial(Block, attn_class=MemEffAttention),
        num_register_tokens=num_register_tokens,
        **kwargs,
    )
    return model


def vit_large(patch_size=16, num_register_tokens=0, **kwargs):
    """Creates a large Vision Transformer model.

    Args:
        patch_size (int, optional): Patch size. Defaults to 16.
        num_register_tokens (int, optional): Number of extra CLS tokens. Defaults to 0.
        **kwargs: Additional keyword arguments.

    Returns:
        DinoVisionTransformer: The constructed model.
    """
    model = DinoVisionTransformer(
        patch_size=patch_size,
        embed_dim=1024,
        depth=24,
        num_heads=16,
        mlp_ratio=4,
        block_fn=partial(Block, attn_class=MemEffAttention),
        num_register_tokens=num_register_tokens,
        **kwargs,
    )
    return model


def vit_giant2(patch_size=16, num_register_tokens=0, **kwargs):
    """Creates a giant Vision Transformer model.

    Args:
        patch_size (int, optional): Patch size. Defaults to 16.
        num_register_tokens (int, optional): Number of extra CLS tokens. Defaults to 0.
        **kwargs: Additional keyword arguments.

    Returns:
        DinoVisionTransformer: The constructed model.
    """
    model = DinoVisionTransformer(
        patch_size=patch_size,
        embed_dim=1536,
        depth=40,
        num_heads=24,
        mlp_ratio=4,
        block_fn=partial(Block, attn_class=MemEffAttention),
        num_register_tokens=num_register_tokens,
        **kwargs,
    )
    return model



# Copyright (c) Meta Platforms, Inc. and affiliates.
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.

import logging
from typing import Dict, Optional

import dinov2.distributed as distributed
import torch
from dinov2.data import DatasetWithEnumeratedTargets, SamplerType, make_data_loader
from dinov2.logging import MetricLogger
from torch import nn
from torchmetrics import MetricCollection

logger = logging.getLogger("dinov2")


class ModelWithNormalize(torch.nn.Module):
    """Model wrapper that normalizes the output of the given model.

    Args:
        model (nn.Module): The model to be wrapped.
    """

    def __init__(self, model):
        super().__init__()
        self.model = model

    def forward(self, samples):
        """Forward pass through the model with normalization.

        Args:
            samples (torch.Tensor): Input samples to the model.

        Returns:
            torch.Tensor: Normalized output from the model.
        """
        return nn.functional.normalize(self.model(samples), dim=1, p=2)


class ModelWithIntermediateLayers(nn.Module):
    """Model that extracts features from intermediate layers.

    Args:
        feature_model (nn.Module): The feature extraction model.
        n_last_blocks (int): Number of last blocks to extract features from.
        autocast_ctx: Autocast context for mixed precision.
    """

    def __init__(self, feature_model, n_last_blocks, autocast_ctx):
        super().__init__()
        self.feature_model = feature_model
        self.feature_model.eval()
        self.n_last_blocks = n_last_blocks
        self.autocast_ctx = autocast_ctx

    def forward(self, images):
        """Forward pass to get features from intermediate layers.

        Args:
            images (torch.Tensor): Input images to the model.

        Returns:
            torch.Tensor: Extracted features from the specified layers.
        """
        with torch.inference_mode():
            with self.autocast_ctx():
                features = self.feature_model.get_intermediate_layers(
                    images, self.n_last_blocks, return_class_token=True
                )
        return features


@torch.inference_mode()
def evaluate(
    model: nn.Module,
    data_loader,
    postprocessors: Dict[str, nn.Module],
    metrics: Dict[str, MetricCollection],
    device: torch.device,
    criterion: Optional[nn.Module] = None,
):
    """Evaluate the model on the given data loader.

    Args:
        model (nn.Module): The model to evaluate.
        data_loader: Data loader for the evaluation dataset.
        postprocessors (Dict[str, nn.Module]): Postprocessing modules for metrics.
        metrics (Dict[str, MetricCollection]): Metrics to compute.
        device (torch.device): Device to perform evaluation on.
        criterion (Optional[nn.Module]): Loss function for evaluation (default is None).

    Returns:
        Tuple[Dict[str, float], Dict[str, float]]: A tuple containing the metric logger stats and computed metrics.
    """
    model.eval()
    if criterion is not None:
        criterion.eval()

    for metric in metrics.values():
        metric = metric.to(device)

    metric_logger = MetricLogger(delimiter="  ")
    header = "Test:"

    for samples, targets, *_ in metric_logger.log_every(data_loader, 10, header):
        outputs = model(samples.to(device))
        targets = targets.to(device)

        if criterion is not None:
            loss = criterion(outputs, targets)
            metric_logger.update(loss=loss.item())

        for k, metric in metrics.items():
            metric_inputs = postprocessors[k](outputs, targets)
            metric.update(**metric_inputs)

    metric_logger.synchronize_between_processes()
    logger.info(f"Averaged stats: {metric_logger}")

    stats = {k: metric.compute() for k, metric in metrics.items()}
    metric_logger_stats = {k: meter.global_avg for k, meter in metric_logger.meters.items()}
    return metric_logger_stats, stats


def all_gather_and_flatten(tensor_rank):
    """Gather tensors from all ranks and flatten the result.

    Args:
        tensor_rank (torch.Tensor): The tensor to gather from all ranks.

    Returns:
        torch.Tensor: Flattened tensor containing data from all ranks.
    """
    tensor_all_ranks = torch.empty(
        distributed.get_global_size(),
        *tensor_rank.shape,
        dtype=tensor_rank.dtype,
        device=tensor_rank.device,
    )
    tensor_list = list(tensor_all_ranks.unbind(0))
    torch.distributed.all_gather(tensor_list, tensor_rank.contiguous())
    return tensor_all_ranks.flatten(end_dim=1)


def extract_features(model, dataset, batch_size, num_workers, gather_on_cpu=False):
    """Extract features from the dataset using the specified model.

    Args:
        model (nn.Module): The model to extract features from.
        dataset: The dataset to extract features from.
        batch_size (int): Batch size for data loading.
        num_workers (int): Number of workers for data loading.
        gather_on_cpu (bool, optional): Whether to gather results on CPU (default is False).

    Returns:
        Tuple[torch.Tensor, torch.Tensor]: Extracted features and corresponding labels.
    """
    dataset_with_enumerated_targets = DatasetWithEnumeratedTargets(dataset)
    sample_count = len(dataset_with_enumerated_targets)
    data_loader = make_data_loader(
        dataset=dataset_with_enumerated_targets,
        batch_size=batch_size,
        num_workers=num_workers,
        sampler_type=SamplerType.DISTRIBUTED,
        drop_last=False,
        shuffle=False,
    )
    return extract_features_with_dataloader(model, data_loader, sample_count, gather_on_cpu)


@torch.inference_mode()
def extract_features_with_dataloader(model, data_loader, sample_count, gather_on_cpu=False):
    """Extract features using a data loader.

    Args:
        model (nn.Module): The model to extract features from.
        data_loader: Data loader for the dataset.
        sample_count (int): Total number of samples in the dataset.
        gather_on_cpu (bool, optional): Whether to gather results on CPU (default is False).

    Returns:
        Tuple[torch.Tensor, torch.Tensor]: Extracted features and corresponding labels.
    """
    gather_device = torch.device("cpu") if gather_on_cpu else torch.device("cuda")
    metric_logger = MetricLogger(delimiter="  ")
    features, all_labels = None, None
    for samples, (index, labels_rank) in metric_logger.log_every(data_loader, 10):
        samples = samples.cuda(non_blocking=True)
        labels_rank = labels_rank.cuda(non_blocking=True)
        index = index.cuda(non_blocking=True)
        features_rank = model(samples).float()

        # init storage feature matrix
        if features is None:
            features = torch.zeros(sample_count, features_rank.shape[-1], device=gather_device)
            labels_shape = list(labels_rank.shape)
            labels_shape[0] = sample_count
            all_labels = torch.full(labels_shape, fill_value=-1, device=gather_device)
            logger.info(f"Storing features into tensor of shape {features.shape}")

        # share indexes, features and labels between processes
        index_all = all_gather_and_flatten(index).to(gather_device)
        features_all_ranks = all_gather_and_flatten(features_rank).to(gather_device)
        labels_all_ranks = all_gather_and_flatten(labels_rank).to(gather_device)

        # update storage feature matrix
        if len(index_all) > 0:
            features.index_copy_(0, index_all, features_all_ranks)
            all_labels.index_copy_(0, index_all, labels_all_ranks)

    logger.info(f"Features shape: {tuple(features.shape)}")
    logger.info(f"Labels shape: {tuple(all_labels.shape)}")

    assert torch.all(all_labels > -1)

    return features, all_labels



# Copyright (c) Facebook, Inc. and its affiliates. (http://www.facebook.com)
# -*- coding: utf-8 -*-
"""mapillary.controllers.detection
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

This module implements the detection based business logic functionalities of the Mapillary
Python SDK.

For more information, please check out https://www.mapillary.com/developer/api-documentation/

- Copyright: (c) 2021 Facebook
- License: MIT LICENSE
"""

# Package imports
import typing

# # Adapter Imports
from zensvi.download.mapillary.models.api.entities import EntityAdapter

# # Models
from zensvi.download.mapillary.models.geojson import GeoJSON

# # Rules
from zensvi.download.mapillary.utils.verify import valid_id

# Local imports


def get_image_detections_controller(image_id: typing.Union[str, int], fields: list = []) -> GeoJSON:
    """Get image detections with given (image) key.

    Args:
        image_id (str): The image id
        fields (list): The fields possible for the detection endpoint.
            Please see https://www.mapillary.com/developer/api-
            documentation for more information

    https://www.mapillary.com/developer/api-documentation

    Returns:
        dict: GeoJSON
    """
    # Checks if the Id given is indeed a valid image_id
    valid_id(identity=image_id, image=True)

    # Return results from the Adapter
    return GeoJSON(
        geojson=EntityAdapter().fetch_detections(
            identity=image_id,
            id_type=True,
            fields=fields,
        )
    )


def get_map_feature_detections_controller(map_feature_id: typing.Union[str, int], fields: list = []) -> GeoJSON:
    """Get image detections with given (map feature) key.

    Args:
        map_feature_id (str): The map feature id
        fields (list): The fields possible for the detection endpoint.
            Please see https://www.mapillary.com/developer/api-
            documentation for more information

    https://www.mapillary.com/developer/api-documentation

    Returns:
        dict: GeoJSON
    """
    # Checks if the Id given is indeed a valid image_id
    valid_id(identity=map_feature_id, image=False)

    # Return results from the Adapter
    return GeoJSON(
        geojson=EntityAdapter().fetch_detections(
            identity=map_feature_id,
            id_type=False,
            fields=fields,
        )
    )



import os

from huggingface_hub import HfApi, hf_hub_download

from zensvi.download.base import BaseDownloader
from zensvi.utils.log import Logger


class GSDownloader(BaseDownloader):
    """Global Streetscapes class to download the original NUS Global Streetscapes
    dataset.

    To download the models used in the dataset, please see
    `cv/classification/utils/global_streetscapes.py`
    """

    def __init__(self, log_path=None):
        super().__init__(log_path)
        self._repo_id = "NUS-UAL/global-streetscapes"
        self._repo_type = "dataset"

        # initialize the logger
        if log_path is not None:
            self.logger = Logger(log_path)
        else:
            self.logger = None

    def _filter_pids_date(self):
        """Required abstract methods from parent class."""
        pass

    def download_svi(self):
        """Required abstract methods from parent class."""
        pass

    @property
    def repo_id(self):
        """Property for Huggingface download.

        :return: repo_id
        :rtype: str
        """
        return self._repo_id

    @property
    def repo_type(self):
        """Property for Huggingface download.

        :return: repo_type
        :rtype: str
        """
        return self._repo_type

    def _download_folder(self, folder_path, local_dir):
        """Download an entire folder from a huggingface dataset repository.

        :param folder_path: Folder path within the repository
        :type folder_path: str
        :param local_dir: Local folder path to download the data into
        :type local_dir: str
        """
        api = HfApi()

        os.makedirs(local_dir, exist_ok=True)

        # list all files in the repo, keep the ones within folder_path
        all_files = api.list_repo_files(self.repo_id, repo_type=self.repo_type)
        files_list = [f for f in all_files if f.startswith(folder_path)]

        # download each of those files
        for file_path in files_list:
            hf_hub_download(repo_id=self.repo_id, repo_type=self.repo_type, filename=file_path, local_dir=local_dir)

    def download_all_data(self, local_dir="data/"):
        """Download all folders and files, recursively, from `data/` This folder
        contains all metadata (csv) for all images of the Global Streetscapes dataset.

        :param local_dir: Local folder to download the data
        :type local_dir: str
        """
        self._download_folder("data/", local_dir)

    def download_manual_labels(self, local_dir="manual_labels/"):
        """Download all folders and files, recursively, from `manual_labels/` This
        folder contains all the manual labels (csv) for train and test images as well as
        the raw images compressed in in tar.gz.

        :param local_dir: Local folder to download the data
        :type local_dir: str
        """
        self._download_folder("manual_labels/", local_dir)

    def download_train(self, local_dir="manual_labels/train/"):
        """Download all folders and files, recursively, from `manual_labels/train/` This
        folder contains all the manual labels (csv) for only the train images.

        :param local_dir: Local folder to download the data
        :type local_dir: str
        """
        self._download_folder("manual_labels/train/", local_dir)

    def download_test(self, local_dir="manual_labels/test/"):
        """Download all folders and files, recursively, from `manual_labels/test/` This
        folder contains all the manual labels (csv) for only the test images.

        :param local_dir: Local folder to download the data
        :type local_dir: str
        """
        self._download_folder("manual_labels/test/", local_dir)

    def download_img_tar(self, local_dir="manual_labels/img/"):
        """Download all folders and files, recursively, from `manual_labels/img/` This
        folder contains all the raw images, compressed in tar.gz.

        :param local_dir: Local folder to download the data
        :type local_dir: str
        """
        self._download_folder("manual_labels/img/", local_dir)



# Copyright (c) Meta Platforms, Inc. and affiliates.
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.

import pathlib

from omegaconf import OmegaConf


def load_config(config_name: str):
    """Load a configuration file.

    Args:
        config_name (str): The name of the configuration file (without the .yaml extension).

    Returns:
        OmegaConf: The loaded configuration as an OmegaConf object.
    """
    config_filename = config_name + ".yaml"
    return OmegaConf.load(pathlib.Path(__file__).parent.resolve() / config_filename)


dinov2_default_config = load_config("ssl_default_config")


def load_and_merge_config(config_name: str):
    """Load a configuration file and merge it with the default configuration.

    Args:
        config_name (str): The name of the configuration file (without the .yaml extension).

    Returns:
        OmegaConf: The merged configuration as an OmegaConf object.
    """
    default_config = OmegaConf.create(dinov2_default_config)
    loaded_config = load_config(config_name)
    return OmegaConf.merge(default_config, loaded_config)



# Copyright (c) Meta Platforms, Inc. and affiliates.
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.

import os
import random
import re
import socket
from typing import Dict, List

import torch
import torch.distributed as dist

_LOCAL_RANK = -1
_LOCAL_WORLD_SIZE = -1


def is_enabled() -> bool:
    """Check if distributed training is enabled.

    Returns:
        bool: True if distributed training is enabled, False otherwise.
    """
    return dist.is_available() and dist.is_initialized()


def get_global_size() -> int:
    """Get the number of processes in the global process group.

    Returns:
        int: The number of processes in the process group.
    """
    return dist.get_world_size() if is_enabled() else 1


def get_global_rank() -> int:
    """Get the rank of the current process within the global process group.

    Returns:
        int: The rank of the current process within the global process group.
    """
    return dist.get_rank() if is_enabled() else 0


def get_local_rank() -> int:
    """Get the rank of the current process within the local (per-machine) process group.

    Returns:
        int: The rank of the current process within the local process group.
    """
    if not is_enabled():
        return 0
    assert 0 <= _LOCAL_RANK < _LOCAL_WORLD_SIZE
    return _LOCAL_RANK


def get_local_size() -> int:
    """Get the size of the per-machine process group.

    Returns:
        int: The size of the per-machine process group, i.e. the number of processes per machine.
    """
    if not is_enabled():
        return 1
    assert 0 <= _LOCAL_RANK < _LOCAL_WORLD_SIZE
    return _LOCAL_WORLD_SIZE


def is_main_process() -> bool:
    """Check if the current process is the main one.

    Returns:
        bool: True if the current process is the main one, False otherwise.
    """
    return get_global_rank() == 0


def _restrict_print_to_main_process() -> None:
    """Disable printing when not in the main process."""
    import builtins as __builtin__

    builtin_print = __builtin__.print

    def print(*args, **kwargs):
        """Custom print function that restricts output to the main process.

        Args:
            *args: Positional arguments to print.
            **kwargs: Keyword arguments to print.
        """
        force = kwargs.pop("force", False)
        if is_main_process() or force:
            builtin_print(*args, **kwargs)

    __builtin__.print = print


def _get_master_port(seed: int = 0) -> int:
    """Get a master port for distributed training.

    Args:
        seed (int, optional): Seed for random number generation. Defaults to 0.

    Returns:
        int: A valid master port number.
    """
    MIN_MASTER_PORT, MAX_MASTER_PORT = (20_000, 60_000)

    master_port_str = os.environ.get("MASTER_PORT")
    if master_port_str is None:
        rng = random.Random(seed)
        return rng.randint(MIN_MASTER_PORT, MAX_MASTER_PORT)

    return int(master_port_str)


def _get_available_port() -> int:
    """Get an available port for binding.

    Returns:
        int: An available port number.
    """
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        # A "" host address means INADDR_ANY i.e. binding to all interfaces.
        # Note this is not compatible with IPv6.
        s.bind(("", 0))
        port = s.getsockname()[1]
        return port


_TORCH_DISTRIBUTED_ENV_VARS = (
    "MASTER_ADDR",
    "MASTER_PORT",
    "RANK",
    "WORLD_SIZE",
    "LOCAL_RANK",
    "LOCAL_WORLD_SIZE",
)


def _collect_env_vars() -> Dict[str, str]:
    """Collect environment variables related to PyTorch distributed training.

    Returns:
        Dict[str, str]: A dictionary of collected environment variables.
    """
    return {env_var: os.environ[env_var] for env_var in _TORCH_DISTRIBUTED_ENV_VARS if env_var in os.environ}


def _is_slurm_job_process() -> bool:
    """Check if the current process is part of a SLURM job.

    Returns:
        bool: True if the current process is part of a SLURM job, False otherwise.
    """
    return "SLURM_JOB_ID" in os.environ


def _parse_slurm_node_list(s: str) -> List[str]:
    """Parse the SLURM node list string into a list of node names.

    Args:
        s (str): The SLURM node list string.

    Returns:
        List[str]: A list of parsed node names.
    """
    nodes = []
    # Extract "hostname", "hostname[1-2,3,4-5]," substrings
    p = re.compile(r"(([^\[]+)(?:\[([^\]]+)\])?),?")
    for m in p.finditer(s):
        prefix, suffixes = s[m.start(2) : m.end(2)], s[m.start(3) : m.end(3)]
        for suffix in suffixes.split(","):
            span = suffix.split("-")
            if len(span) == 1:
                nodes.append(prefix + suffix)
            else:
                width = len(span[0])
                start, end = int(span[0]), int(span[1]) + 1
                nodes.extend([prefix + f"{i:0{width}}" for i in range(start, end)])
    return nodes


def _check_env_variable(key: str, new_value: str):
    """Check if an environment variable can be set to a new value.

    Args:
        key (str): The name of the environment variable.
        new_value (str): The new value to set.

    Raises:
        RuntimeError: If the environment variable is already set to a different value.
    """
    # Only check for difference with preset environment variables
    if key in os.environ and os.environ[key] != new_value:
        raise RuntimeError(f"Cannot export environment variables as {key} is already set")


class _TorchDistributedEnvironment:
    """Class to manage the PyTorch distributed environment."""

    def __init__(self):
        """Initialize the distributed environment settings."""
        self.master_addr = "127.0.0.1"
        self.master_port = 0
        self.rank = -1
        self.world_size = -1
        self.local_rank = -1
        self.local_world_size = -1

        if _is_slurm_job_process():
            return self._set_from_slurm_env()

        env_vars = _collect_env_vars()
        if not env_vars:
            # Environment is not set
            pass
        elif len(env_vars) == len(_TORCH_DISTRIBUTED_ENV_VARS):
            # Environment is fully set
            return self._set_from_preset_env()
        else:
            # Environment is partially set
            collected_env_vars = ", ".join(env_vars.keys())
            raise RuntimeError(f"Partially set environment: {collected_env_vars}")

        if torch.cuda.device_count() > 0:
            return self._set_from_local()

        raise RuntimeError("Can't initialize PyTorch distributed environment")

    def _set_from_slurm_env(self):
        """Set environment variables from SLURM job environment."""
        # logger.info("Initialization from Slurm environment")
        job_id = int(os.environ["SLURM_JOB_ID"])
        node_count = int(os.environ["SLURM_JOB_NUM_NODES"])
        nodes = _parse_slurm_node_list(os.environ["SLURM_JOB_NODELIST"])
        assert len(nodes) == node_count

        self.master_addr = nodes[0]
        self.master_port = _get_master_port(seed=job_id)
        self.rank = int(os.environ["SLURM_PROCID"])
        self.world_size = int(os.environ["SLURM_NTASKS"])
        assert self.rank < self.world_size
        self.local_rank = int(os.environ["SLURM_LOCALID"])
        self.local_world_size = self.world_size // node_count
        assert self.local_rank < self.local_world_size

    def _set_from_preset_env(self):
        """Set environment variables from preset environment (i.e. torchrun)."""
        # logger.info("Initialization from preset environment")
        self.master_addr = os.environ["MASTER_ADDR"]
        self.master_port = os.environ["MASTER_PORT"]
        self.rank = int(os.environ["RANK"])
        self.world_size = int(os.environ["WORLD_SIZE"])
        assert self.rank < self.world_size
        self.local_rank = int(os.environ["LOCAL_RANK"])
        self.local_world_size = int(os.environ["LOCAL_WORLD_SIZE"])
        assert self.local_rank < self.local_world_size

    def _set_from_local(self):
        """Set environment variables for a single node and GPU job (i.e. local script run)."""
        # logger.info("Initialization from local")
        self.master_addr = "127.0.0.1"
        self.master_port = _get_available_port()
        self.rank = 0
        self.world_size = 1
        self.local_rank = 0
        self.local_world_size = 1

    def export(self, *, overwrite: bool) -> "_TorchDistributedEnvironment":
        """Export the environment variables for distributed training.

        Args:
            overwrite (bool): If True, overwrites already set variables. Else fails.

        Returns:
            _TorchDistributedEnvironment: The current instance of the distributed environment.
        """
        # See the "Environment variable initialization" section from
        # https://pytorch.org/docs/stable/distributed.html for the complete list of
        # environment variables required for the env:// initialization method.
        env_vars = {
            "MASTER_ADDR": self.master_addr,
            "MASTER_PORT": str(self.master_port),
            "RANK": str(self.rank),
            "WORLD_SIZE": str(self.world_size),
            "LOCAL_RANK": str(self.local_rank),
            "LOCAL_WORLD_SIZE": str(self.local_world_size),
        }
        if not overwrite:
            for k, v in env_vars.items():
                _check_env_variable(k, v)

        os.environ.update(env_vars)
        return self


def enable(
    *,
    set_cuda_current_device: bool = True,
    overwrite: bool = False,
    allow_nccl_timeout: bool = False,
):
    """Enable distributed mode.

    Args:
        set_cuda_current_device (bool, optional): If True, call torch.cuda.set_device() to set the
            current PyTorch CUDA device to the one matching the local rank. Defaults to True.
        overwrite (bool, optional): If True, overwrites already set variables. Else fails. Defaults to False.
        allow_nccl_timeout (bool, optional): If True, allows NCCL timeout. Defaults to False.

    Raises:
        RuntimeError: If distributed mode has already been enabled.
    """
    global _LOCAL_RANK, _LOCAL_WORLD_SIZE
    if _LOCAL_RANK >= 0 or _LOCAL_WORLD_SIZE >= 0:
        raise RuntimeError("Distributed mode has already been enabled")
    torch_env = _TorchDistributedEnvironment()
    torch_env.export(overwrite=overwrite)

    if set_cuda_current_device:
        torch.cuda.set_device(torch_env.local_rank)

    if allow_nccl_timeout:
        # This allows to use torch distributed timeout in a NCCL backend
        key, value = "NCCL_ASYNC_ERROR_HANDLING", "1"
        if not overwrite:
            _check_env_variable(key, value)
        os.environ[key] = value

    dist.init_process_group(backend="nccl")
    dist.barrier()

    # Finalize setup
    _LOCAL_RANK = torch_env.local_rank
    _LOCAL_WORLD_SIZE = torch_env.local_world_size
    _restrict_print_to_main_process()



# Copyright (c) Meta Platforms, Inc. and affiliates.
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.

from io import BytesIO
from typing import Any

from PIL import Image


class Decoder:
    """Abstract base class for decoders.

    This class defines the interface for decoding data.
    """

    def decode(self) -> Any:
        """Decode the data.

        Raises:
            NotImplementedError: If the decode method is not implemented.
        """
        raise NotImplementedError


class ImageDataDecoder(Decoder):
    """Decoder for image data.

    This class is responsible for decoding image data from bytes.

    Args:
        image_data (bytes): The image data in bytes format.
    """

    def __init__(self, image_data: bytes) -> None:
        self._image_data = image_data

    def decode(self) -> Image:
        """Decode the image data to a PIL Image.

        Returns:
            Image: The decoded image in RGB format.
        """
        f = BytesIO(self._image_data)
        return Image.open(f).convert(mode="RGB")


class TargetDecoder(Decoder):
    """Decoder for target data.

    This class is responsible for returning the target data.

    Args:
        target (Any): The target data to be decoded.
    """

    def __init__(self, target: Any):
        self._target = target

    def decode(self) -> Any:
        """Return the target data.

        Returns:
            Any: The target data.
        """
        return self._target



# MIT License

# Copyright (c) 2022 Intelligent Systems Lab Org

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# File author: Shariq Farooq Bhat

import torch
import torch.nn as nn


class PatchTransformerEncoder(nn.Module):
    """Patch Transformer Encoder.

    This class implements a ViT-like transformer block that processes input
    features through a series of transformer encoder layers.

    Attributes:
        use_class_token (bool): Indicates whether to use a class token.
        transformer_encoder (nn.TransformerEncoder): The transformer encoder.
        embedding_convPxP (nn.Conv2d): Convolutional layer for embedding.
    """

    def __init__(
        self,
        in_channels,
        patch_size=10,
        embedding_dim=128,
        num_heads=4,
        use_class_token=False,
    ):
        """Initializes the PatchTransformerEncoder.

        Args:
            in_channels (int): Number of input channels.
            patch_size (int, optional): Size of the patches. Defaults to 10.
            embedding_dim (int, optional): Dimension of the embedding. Defaults to 128.
            num_heads (int, optional): Number of attention heads. Defaults to 4.
            use_class_token (bool, optional): Whether to use a class token. Defaults to False.
        """
        super(PatchTransformerEncoder, self).__init__()
        self.use_class_token = use_class_token
        encoder_layers = nn.TransformerEncoderLayer(embedding_dim, num_heads, dim_feedforward=1024)
        self.transformer_encoder = nn.TransformerEncoder(encoder_layers, num_layers=4)  # takes shape S,N,E

        self.embedding_convPxP = nn.Conv2d(
            in_channels,
            embedding_dim,
            kernel_size=patch_size,
            stride=patch_size,
            padding=0,
        )

    def positional_encoding_1d(self, sequence_length, batch_size, embedding_dim, device="cpu"):
        """Generates positional encodings.

        Args:
            sequence_length (int): Length of the sequence.
            batch_size (int): Size of the batch.
            embedding_dim (int): Dimension of the embedding.
            device (str, optional): Device to perform the computation on. Defaults to "cpu".

        Returns:
            torch.Tensor: Positional encodings of shape (S, B, E).
        """
        position = torch.arange(0, sequence_length, dtype=torch.float32, device=device).unsqueeze(1)
        index = torch.arange(0, embedding_dim, 2, dtype=torch.float32, device=device).unsqueeze(0)
        div_term = torch.exp(index * (-torch.log(torch.tensor(10000.0, device=device)) / embedding_dim))
        pos_encoding = position * div_term
        pos_encoding = torch.cat([torch.sin(pos_encoding), torch.cos(pos_encoding)], dim=1)
        pos_encoding = pos_encoding.unsqueeze(1).repeat(1, batch_size, 1)
        return pos_encoding

    def forward(self, x):
        """Performs the forward pass.

        Args:
            x (torch.Tensor): Input feature tensor of shape (N, C, H, W).

        Returns:
            torch.Tensor: Transformer output embeddings of shape (S, N, E),
            where S is the sequence length, N is the batch size, and E is the embedding dimension.
        """
        embeddings = self.embedding_convPxP(x).flatten(2)  # .shape = n,c,s = n, embedding_dim, s
        if self.use_class_token:
            # extra special token at start ?
            embeddings = nn.functional.pad(embeddings, (1, 0))

        # change to S,N,E format required by transformer
        embeddings = embeddings.permute(2, 0, 1)
        S, N, E = embeddings.shape
        embeddings = embeddings + self.positional_encoding_1d(S, N, E, device=embeddings.device)
        x = self.transformer_encoder(embeddings)  # .shape = S, N, E
        return x



from .glare import ClassifierGlare
from .lighting import ClassifierLighting
from .panorama import ClassifierPanorama
from .perception import ClassifierPerception, ClassifierPerceptionViT
from .places365 import ClassifierPlaces365
from .platform import ClassifierPlatform
from .quality import ClassifierQuality
from .reflection import ClassifierReflection
from .view_direction import ClassifierViewDirection
from .weather import ClassifierWeather

__all__ = [
    "ClassifierGlare",
    "ClassifierLighting",
    "ClassifierPanorama",
    "ClassifierPerception",
    "ClassifierPerceptionViT",
    "ClassifierPlaces365",
    "ClassifierPlatform",
    "ClassifierQuality",
    "ClassifierReflection",
    "ClassifierViewDirection",
    "ClassifierWeather",
]



import datetime
import glob
import logging
import os
import random
import warnings
from concurrent.futures import ThreadPoolExecutor, as_completed
from pathlib import Path

import geopandas as gpd
import osmnx as ox
import pandas as pd
import requests
from PIL import Image
from shapely.errors import ShapelyDeprecationWarning
from tqdm import tqdm

import zensvi.download.kartaview.download_functions as kv
from zensvi.download.base import BaseDownloader
from zensvi.download.utils.helpers import check_and_buffer, standardize_column_names
from zensvi.utils.log import Logger

warnings.filterwarnings("ignore", category=ShapelyDeprecationWarning)
logging.getLogger("mapillary.utils.client").setLevel(logging.WARNING)


class KVDownloader(BaseDownloader):
    """KartaView Downloader class.

    Args:
        log_path (str, optional): Path to the log file. Defaults to None.
        max_workers (int, optional): Number of workers for parallel processing. Defaults to None.
    """

    def __init__(self, log_path=None, max_workers=None):
        super().__init__(log_path)
        self._max_workers = max_workers
        # initialize the logger
        if log_path is not None:
            # make log directory
            Path(log_path).parent.mkdir(parents=True, exist_ok=True)
            self.logger = Logger(log_path)
        else:
            self.logger = None

    @property
    def max_workers(self):
        """Property for the number of workers for parallel processing.

        Returns:
            int: max_workers
        """
        return self._max_workers

    @max_workers.setter
    def max_workers(self, max_workers):
        if max_workers is None:
            self._max_workers = min(32, os.cpu_count() + 4)
        else:
            self._max_workers = max_workers

    def _read_pids(self, path_pid):
        pid_df = pd.read_csv(path_pid)
        # drop NA values in id columns
        pid_df = pid_df.dropna(subset=["id"])
        # get unique pids (ie "id" columns) as a list
        pids = pid_df["id"].astype("int64").unique().tolist()
        return pids

    def _set_dirs(self, dir_output):
        # set dir_output as attribute and create the directory
        self.dir_output = Path(dir_output)
        self.dir_output.mkdir(parents=True, exist_ok=True)
        self.pids_url = self.dir_output / "pids_urls_kv.csv"
        # # set dir_cache as attribute and create the directory
        # self.dir_cache = self.dir_output / "cache_zensvi_kv"
        # self.dir_cache.mkdir(parents=True, exist_ok=True)
        # # set other cache directories
        # self.cache_lat_lon = self.dir_cache / "lat_lon_kv.csv"
        # self.cache_pids_raw = self.dir_cache / "pids_raw_kv.csv"

    def _get_pids_from_gdf(self, gdf):
        # set crs to EPSG:4326 if it's None
        if gdf.crs is None:
            gdf = gdf.set_crs("EPSG:4326")
        elif gdf.crs != "EPSG:4326":
            # convert to EPSG:4326
            gdf = gdf.to_crs("EPSG:4326")
        # use get pids with kv.get_points_in_shape
        result_df = kv.get_points_in_shape(gdf)
        return result_df

    def _get_raw_pids(self, **kwargs):
        # if self.cache_pids_raw.exists():
        #     pid = pd.read_csv(self.cache_pids_raw)
        #     print("The raw image IDs have been read from the cache")
        #     return pid

        # input: lat and lon
        if kwargs["lat"] is not None and kwargs["lon"] is not None:
            # create a geodataframe with lat and lon
            df = pd.DataFrame({"lat": [kwargs["lat"]], "lon": [kwargs["lon"]]})
            gdf = gpd.GeoDataFrame(df, geometry=gpd.points_from_xy(df.lon, df.lat), crs="EPSG:4326")

        # input: csv file
        elif kwargs["input_csv_file"] != "":
            df = pd.read_csv(kwargs["input_csv_file"])
            df = standardize_column_names(df)
            gdf = gpd.GeoDataFrame(
                df,
                geometry=gpd.points_from_xy(df.longitude, df.latitude),
                crs="EPSG:4326",
            )

        # input: shapefile
        elif kwargs["input_shp_file"] != "":
            gdf = gpd.read_file(kwargs["input_shp_file"])

        # input: place name
        elif kwargs["input_place_name"] != "":
            print("Geocoding the input place name")
            gdf = ox.geocoder.geocode_to_gdf(kwargs["input_place_name"])
            # raise error if the input_place_name is not found
            if len(gdf) == 0:
                raise ValueError("The input_place_name is not found. Please try another place name.")
        else:
            raise ValueError("Please input the lat and lon, csv file, or shapefile.")

        # check geometry type and buffer
        gdf = check_and_buffer(gdf, kwargs["buffer"])
        # get pid
        pid = self._get_pids_from_gdf(gdf)
        pid = self._filter_pids_date(pid, kwargs["start_date"], kwargs["end_date"])

        return pid

    def _filter_pids_date(self, pid_df, start_date, end_date):
        # create a temporary column date from captured_at (milliseconds from Unix epoch)
        pid_df["date"] = pd.to_datetime(pid_df["shotDate"], format="%Y-%m-%d %H:%M:%S")
        # check if start_date and end_date are in the correct format with regex. If not, raise error
        if start_date is not None:
            try:
                start_date = datetime.datetime.strptime(start_date, "%Y-%m-%d")
            except ValueError:
                raise ValueError("Incorrect start_date format, should be YYYY-MM-DD")
        if end_date is not None:
            try:
                end_date = datetime.datetime.strptime(end_date, "%Y-%m-%d")
            except ValueError:
                raise ValueError("Incorrect end_date format, should be YYYY-MM-DD")
        # if start_date is not None, filter out the rows with date < start_date
        pid_df = pid_df[pid_df["date"] >= start_date] if start_date is not None else pid_df
        # if end_date is not None, filter out the rows with date > end_date
        pid_df = pid_df[pid_df["date"] <= end_date] if end_date is not None else pid_df
        # drop the temporary column date
        pid_df = pid_df.drop(columns="date")
        return pid_df

    def _get_pids(self, path_pid, **kwargs):
        # get raw pid
        pid = self._get_raw_pids(**kwargs)

        if pid is None:
            print("There is no image ID to download")
            return

        # move the "id" column to the first column
        pid = pid[["id"] + [col for col in pid.columns if col != "id"]]

        pid.to_csv(path_pid, index=False)
        print("The image IDs have been saved to {}".format(path_pid))

    def _get_urls_kv(self, path_pid):

        pid = pd.read_csv(path_pid)
        if len(pid) == 0:
            print("There is no image ID to download")
            return
        urls = pid[["id", "fileurlProc"]].rename(columns={"fileurlProc": "url"})
        urls.to_csv(self.pids_url, index=False)

    def _download_images_kv(self, path_pid, cropped, batch_size, start_date, end_date):
        checkpoints = glob.glob(str(self.panorama_output / "**/*.png"), recursive=True)

        # Read already downloaded images and convert to ids
        downloaded_ids = set([int(Path(file_path).stem) for file_path in checkpoints])  # Use set for faster operations

        pid_df = pd.read_csv(path_pid).dropna(subset=["id"])
        pid_df["id"] = pid_df["id"].astype("int64")
        urls_df = pd.read_csv(self.pids_url)
        urls_df["id"] = urls_df["id"].astype("int64")
        # merge pid_df and urls_df
        urls_df = urls_df.merge(pid_df, on="id", how="left")
        # filter out the rows by date
        urls_df = self._filter_pids_date(urls_df, start_date, end_date)

        # Filter out the ids that have already been processed
        urls_df = urls_df[~urls_df["id"].isin(downloaded_ids)]  # Use isin for efficient operation

        def worker(row, output_dir, cropped, max_retries=5):
            url, panoid = row.url, row.id
            user_agent = random.choice(self._user_agents)
            proxy = random.choice(self._proxies)

            image_name = f"{panoid}.png"  # Use id for file name
            image_path = output_dir / image_name

            for retry in range(max_retries):
                try:
                    response = requests.get(url, headers=user_agent, proxies=proxy, timeout=10)
                    if response.status_code == 200:
                        with open(image_path, "wb") as f:
                            f.write(response.content)

                        if cropped:
                            img = Image.open(image_path)
                            w, h = img.size
                            img_cropped = img.crop((0, 0, w, h // 2))
                            img_cropped.save(image_path)
                        break
                    else:
                        if retry == max_retries - 1:  # Last retry
                            if self.logger is not None:
                                self.logger.log_failed_pids(panoid)
                except Exception as e:
                    if retry == max_retries - 1:  # Last retry
                        if self.logger is not None:
                            self.logger.log_failed_pids(panoid)
                        print(f"Error: {e}")
                    else:
                        print(f"Retry {retry + 1}/{max_retries} failed: {e}")

        num_batches = (len(urls_df) + batch_size - 1) // batch_size

        # Calculate current highest batch number
        existing_batches = glob.glob(str(self.panorama_output / "batch_*"))
        existing_batch_numbers = [int(Path(batch).name.split("_")[-1]) for batch in existing_batches]
        start_batch_number = max(existing_batch_numbers, default=0)

        for i in tqdm(
            range(start_batch_number, start_batch_number + num_batches),
            desc=f"Downloading images by batch size {min(batch_size, len(urls_df))}",
        ):
            # Create a new sub-folder for each batch
            batch_out_path = self.panorama_output / f"batch_{i+1}"
            batch_out_path.mkdir(exist_ok=True)

            with ThreadPoolExecutor(max_workers=self.max_workers) as executor:
                batch_futures = {
                    executor.submit(worker, row, batch_out_path, cropped): row.id
                    for row in urls_df.iloc[
                        (i - start_batch_number) * batch_size : (i + 1 - start_batch_number) * batch_size
                    ].itertuples()
                }
                for future in tqdm(
                    as_completed(batch_futures),
                    total=len(batch_futures),
                    desc=f"Downloading images for batch #{i+1}",
                ):
                    try:
                        future.result()
                    except Exception as e:
                        print(f"Error: {e}")

    def download_svi(
        self,
        dir_output,
        path_pid=None,
        lat=None,
        lon=None,
        input_csv_file="",
        input_shp_file="",
        input_place_name="",
        buffer=10,
        update_pids=False,
        cropped=False,
        batch_size=1000,
        start_date=None,
        end_date=None,
        metadata_only=False,
        max_workers=None,
    ):
        """Downloads street view images from KartaView using specified parameters.

        Args:
            dir_output (str): Directory where output files and images will be stored.
            path_pid (str, optional): Path to a file containing panorama IDs. If not provided, IDs will be fetched based on other parameters.
            lat (float, optional): Latitude to fetch panorama IDs around this point. Must be used with `lon`.
            lon (float, optional): Longitude to fetch panorama IDs around this point. Must be used with `lat`.
            input_csv_file (str, optional): Path to a CSV file containing locations for which to fetch panorama IDs.
            input_shp_file (str, optional): Path to a shapefile containing geographic locations for fetching panorama IDs.
            input_place_name (str, optional): A place name for geocoding to fetch panorama IDs.
            buffer (int, optional): Buffer size in meters to expand the geographic area for panorama ID fetching.
            update_pids (bool, optional): If True, will update panorama IDs even if a valid `path_pid` is provided. Defaults to False.
            cropped (bool, optional): If True, images will be cropped to the upper half. Defaults to False.
            batch_size (int, optional): Number of images to process in each batch. Defaults to 1000.
            start_date (str, optional): Start date (YYYY-MM-DD) to filter images by capture date.
            end_date (str, optional): End date (YYYY-MM-DD) to filter images by capture date.
            metadata_only (bool, optional): If True, skips downloading images and only fetches metadata. Defaults to False.
            max_workers (int, optional): Number of workers for parallel processing. If None, it will be set to min(32, os.cpu_count() + 4).

        Returns:
            None: This method does not return a value but will save files directly to the specified output directory.

        Raises:
            ValueError: If required parameters for fetching panorama IDs are not adequately specified.
            FileNotFoundError: If `path_pid` is specified but the file does not exist.

        Notes:
            This method logs significant events and errors, making it suitable for both interactive usage and automated workflows.
        """
        if self.logger is not None:
            # record the arguments
            self.logger.log_args(
                "KVDownloader download_svi",
                dir_output=dir_output,
                path_pid=path_pid,
                lat=lat,
                lon=lon,
                input_csv_file=input_csv_file,
                input_shp_file=input_shp_file,
                input_place_name=input_place_name,
                buffer=buffer,
                update_pids=update_pids,
                cropped=cropped,
                batch_size=batch_size,
                start_date=start_date,
                end_date=end_date,
                metadata_only=metadata_only,
                max_workers=max_workers,
            )
        # Set max_workers
        self.max_workers = max_workers

        # set necessary directories
        self._set_dirs(dir_output)

        # call _get_pids function first if path_pid is None
        if path_pid is None:
            print("Getting pids...")
            path_pid = self.dir_output / "kv_pids.csv"
            if path_pid.exists() & (not update_pids):
                print("update_pids is set to False. So the following csv file will be used: {}".format(path_pid))
            else:
                self._get_pids(
                    path_pid,
                    lat=lat,
                    lon=lon,
                    input_csv_file=input_csv_file,
                    input_shp_file=input_shp_file,
                    input_place_name=input_place_name,
                    buffer=buffer,
                    start_date=start_date,
                    end_date=end_date,
                )
        else:
            # check if the path_pid exists
            if Path(path_pid).exists():
                print("The following csv file will be used: {}".format(path_pid))
            else:
                self._get_pids(
                    path_pid,
                    lat=lat,
                    lon=lon,
                    input_csv_file=input_csv_file,
                    input_shp_file=input_shp_file,
                    input_place_name=input_place_name,
                    buffer=buffer,
                    start_date=start_date,
                    end_date=end_date,
                )
        # stop if metadata_only is True
        if metadata_only:
            print("The metadata has been downloaded")
            return

        # create a folder within self.dir_output
        self.panorama_output = self.dir_output / "kv_svi"
        self.panorama_output.mkdir(parents=True, exist_ok=True)

        # get urls
        if path_pid.exists():
            self._get_urls_kv(path_pid)
            # download images
            self._download_images_kv(path_pid, cropped, batch_size, start_date, end_date)
        else:
            print("There is no imagae ID to download within the given input parameters")

        # # delete the cache directory
        # if self.dir_cache.exists():
        #     shutil.rmtree(self.dir_cache)
        #     print("The cache directory has been deleted")



# Copyright (c) Meta Platforms, Inc. and affiliates.
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.

import logging
import os
import sys

from dinov2.logging import setup_logging
from dinov2.run.submit import get_args_parser, submit_jobs
from dinov2.train import get_args_parser as get_train_args_parser

logger = logging.getLogger("dinov2")


class Trainer(object):
    """Trainer class for managing the training process.

    This class is responsible for setting up the training arguments,
    executing the training process, and handling checkpointing.

    Attributes:
        args: The arguments for training, including configuration and output directory.
    """

    def __init__(self, args):
        """Initializes the Trainer with the given arguments.

        Args:
            args: The arguments for training.
        """
        self.args = args

    def __call__(self):
        """Executes the training process by calling the main training function."""
        from dinov2.train import main as train_main

        self._setup_args()
        train_main(self.args)

    def checkpoint(self):
        """Creates a checkpoint for the current training state.

        Returns:
            A DelayedSubmission object that can be used to requeue the training job.
        """
        import submitit

        logger.info(f"Requeuing {self.args}")
        empty = type(self)(self.args)
        return submitit.helpers.DelayedSubmission(empty)

    def _setup_args(self):
        """Sets up the training arguments, including output directory and logging.

        This method modifies the output directory to include the job ID
        and logs the process group information and the arguments.
        """
        import submitit

        job_env = submitit.JobEnvironment()
        self.args.output_dir = self.args.output_dir.replace("%j", str(job_env.job_id))
        logger.info(f"Process group: {job_env.num_tasks} tasks, rank: {job_env.global_rank}")
        logger.info(f"Args: {self.args}")


def main():
    """Main function to launch the DINOv2 training process.

    This function sets up argument parsing, logging, and submits the training job.

    Returns:
        int: Exit status code.
    """
    description = "Submitit launcher for DINOv2 training"
    train_args_parser = get_train_args_parser(add_help=False)
    parents = [train_args_parser]
    args_parser = get_args_parser(description=description, parents=parents)
    args = args_parser.parse_args()

    setup_logging()

    assert os.path.exists(args.config_file), "Configuration file does not exist!"
    submit_jobs(Trainer, args, name="dinov2:train")
    return 0


if __name__ == "__main__":
    sys.exit(main())



import glob
from pathlib import Path
from typing import Any, Tuple, Union

import contextily as ctx
import geopandas as gpd
import h3
import osmnx as ox
import pandas as pd
from matplotlib import colors
from matplotlib import pyplot as plt
from shapely.geometry import Polygon

from .font_property import _get_font_properties


def _lat_lng_to_h3(row, resolution=7):
    """Convert latitude and longitude to H3 hex ID at the specified resolution."""
    return h3.geo_to_h3(row["lat"], row["lon"], resolution)


def _h3_to_polygon(hex_id):
    """Convert H3 hex ID to a Shapely polygon."""
    vertices = h3.h3_to_geo_boundary(hex_id, geo_json=True)
    return Polygon(vertices)


def _create_line(gdf, variable_name=None):
    # gdf is a point GeoDataFrame, so convert it to polygon by taking convex hull
    polygon = gdf["geometry"].buffer(100).to_crs(4326).unary_union
    # then use osmnx to get street network graph_from_polygon
    G = ox.graph_from_polygon(polygon, network_type="all", retain_all=True)
    # convert to GeoDataFrame
    line_gdf = ox.graph_to_gdfs(G, nodes=False, edges=True).reset_index()
    # attached point data to the nearest edge (within some distance tolerance)
    line_gdf = line_gdf.to_crs(gdf.crs)
    # join point to nearest line segment
    gdf = gpd.sjoin_nearest(gdf, line_gdf, max_distance=100, how="left")
    # left join back to the line_gd
    if variable_name:
        # aggregate by the id and get mean of the variable
        aggregated_data = gdf.groupby("u")[variable_name].mean().reset_index(name="mean_value")
        gdf[variable_name] = gdf["u"].map(aggregated_data.set_index("u")["mean_value"])
    else:
        # aggregate by the id and get count
        aggregated_data = gdf.groupby("u").size().reset_index(name="count")
        gdf["count"] = gdf["u"].map(aggregated_data.set_index("u")["count"])
    # drop geomtry column in gdf
    gdf = gdf.drop(columns="geometry")
    # join gdf to line_gdf
    line_gdf = gpd.GeoDataFrame(line_gdf.merge(gdf, on="u", how="left"))
    return line_gdf.to_crs(3857)


def _create_hexagon(gdf, resolution=7, variable_name=None):
    gdf = gdf.to_crs(4326)
    gdf["h3_id"] = gdf.apply(_lat_lng_to_h3, resolution=resolution, axis=1)
    if variable_name:
        aggregated_data = gdf.groupby("h3_id")[variable_name].mean().reset_index(name="mean_value")
        gdf[variable_name] = gdf["h3_id"].map(aggregated_data.set_index("h3_id")["mean_value"])
    else:
        aggregated_data = gdf.groupby("h3_id").size().reset_index(name="count")
        gdf["count"] = gdf["h3_id"].map(aggregated_data.set_index("h3_id")["count"])
    gdf["geometry"] = gdf["h3_id"].apply(_h3_to_polygon)
    hex_gdf = gpd.GeoDataFrame(gdf, geometry="geometry")
    return hex_gdf.to_crs(3857)


def _add_colorbar(
    fig,
    ax,
    vmin,
    vmax,
    cmap,
    legend_title,
    prop,
    prop_legend,
    font_color,
    orientation="vertical",
    dark_mode=False,
):
    """Adds a colorbar to the figure based on given parameters, with optional dark
    mode.
    """
    norm = colors.Normalize(vmin=vmin, vmax=vmax)
    sm = plt.cm.ScalarMappable(norm=norm, cmap=cmap)
    sm.set_array([])
    cbar = fig.colorbar(sm, ax=ax, orientation=orientation, fraction=0.036, pad=0.04)
    cbar.set_label(legend_title, fontproperties=prop, color=font_color)
    cbar.ax.tick_params(labelsize=prop_legend.get_size(), color=font_color)
    cbar.outline.set_edgecolor(font_color)
    for label in cbar.ax.get_xticklabels():
        label.set_fontproperties(prop_legend)
        label.set_color(font_color)


def plot_map(
    path_pid: Union[str, Path],
    pid_column: str = "panoid",
    dir_input: Union[str, Path] = None,
    csv_file_pattern: str = "*.csv",
    variable_name: str = None,
    plot_type: str = "point",
    path_output: Union[str, Path] = None,
    resolution: int = 7,
    cmap: str = "viridis",
    legend: bool = True,
    title: str = None,
    legend_title: str = None,
    basemap_source: Any = ctx.providers.CartoDB.PositronNoLabels,
    figure_size: Tuple[int, int] = (10, 10),
    dpi: int = 300,
    font_size: int = 30,
    dark_mode: bool = False,
    **kwargs,
) -> Tuple[plt.Figure, plt.Axes]:
    """Plots a geographic map from data points, with options for line or hexagonal
    aggregations, coloring by variables, and using different base maps.

    Args:
        path_pid (Union[str, Path]): Path to the CSV file containing longitude and latitude and other metadata.
        pid_column (str, optional): Column name in CSV that acts as a primary key or identifier. Defaults to "panoid".
        dir_input (Union[str, Path], optional): Directory path where additional CSV data files are stored, matched by pattern. Defaults to None.
        csv_file_pattern (str, optional): Pattern to match CSV files in the directory. Defaults to None.
        variable_name (str, optional): Name of the variable in CSV to use for coloring and aggregation. Defaults to None.
        plot_type (str, optional): Type of plot to generate: 'point', 'line', or 'hexagon'. Defaults to "point".
        path_output (Union[str, Path], optional): Path where the plotted figure will be saved. Defaults to None.
        resolution (int, optional): Resolution level for H3 hexagonal tiling. Defaults to 7.
        cmap (str, optional): Colormap for the plot. Defaults to "viridis".
        legend (bool, optional): Whether to add a color legend to the plot. Defaults to True.
        title (str, optional): Title of the plot. Defaults to None.
        legend_title (str, optional): Title for the legend. Defaults to None.
        basemap_source (Any, optional): Contextily basemap source. Defaults to ctx.providers.CartoDB.PositronNoLabels.
        dpi (int, optional): Dots per inch (resolution) of the output image file. Defaults to 300.
        font_size (int, optional): Font size for titles and legend. Defaults to 30.
        dark_mode (bool, optional): Whether to use a dark theme for the plot. Defaults to False.
        **kwargs: Additional keyword arguments passed to GeoPandas plot function.

    Returns:
        Tuple[plt.Figure, plt.Axes]: A tuple containing the Matplotlib figure and axes objects.

    Raises:
        ValueError: If an invalid `plot_type` is provided.
    """
    font_color = "#2b2b2b" if not dark_mode else "white"
    # Load path_pid with longitude and latitude
    pid_df = pd.read_csv(path_pid)

    fig, ax = plt.subplots(figsize=figure_size)
    if dark_mode:
        plt.style.use("dark_background")
    gdf = None
    if dir_input and csv_file_pattern and variable_name:
        dir_input = Path(dir_input)
        csv_files = glob.glob(str(dir_input / "**" / csv_file_pattern), recursive=True)
        df_list = [pd.read_csv(file) for file in csv_files]
        merged_df = pd.concat(df_list, ignore_index=True)
        final_df = pd.merge(pid_df, merged_df, left_on=pid_column, right_on="filename_key", how="inner")
        gdf = gpd.GeoDataFrame(
            final_df,
            geometry=gpd.points_from_xy(final_df.lon, final_df.lat),
            crs="EPSG:4326",
        )
        gdf = gdf.to_crs(3857)
        if plot_type == "line":
            gdf = _create_line(gdf, variable_name=variable_name)
        elif plot_type == "hexagon":
            gdf = _create_hexagon(gdf, resolution=resolution, variable_name=variable_name)
        gdf.plot(ax=ax, column=variable_name, cmap=cmap, **kwargs)
    elif variable_name:
        gdf = gpd.GeoDataFrame(pid_df, geometry=gpd.points_from_xy(pid_df.lon, pid_df.lat), crs="EPSG:4326")
        gdf = gdf.to_crs(3857)
        if plot_type == "point":
            gdf.plot(ax=ax, column=variable_name, cmap=cmap, **kwargs)
        elif plot_type == "line":
            gdf = _create_line(gdf, variable_name=variable_name)
            gdf.plot(ax=ax, column=variable_name, cmap=cmap, **kwargs)
        elif plot_type == "hexagon":
            gdf = _create_hexagon(gdf, resolution=resolution, variable_name=variable_name)
            gdf.plot(ax=ax, column=variable_name, cmap=cmap, **kwargs)
    else:
        gdf = gpd.GeoDataFrame(pid_df, geometry=gpd.points_from_xy(pid_df.lon, pid_df.lat), crs="EPSG:4326")
        gdf = gdf.to_crs(3857)
        if plot_type == "point":
            gdf.plot(ax=ax, **kwargs)
        elif plot_type == "line":
            gdf = _create_line(gdf)
            gdf.plot(ax=ax, column="count", cmap=cmap, **kwargs)
        elif plot_type == "hexagon":
            gdf = _create_hexagon(gdf, resolution=resolution)
            gdf.plot(ax=ax, column="count", cmap=cmap, **kwargs)
        else:
            raise ValueError("Invalid plot type")
    # Add basemap if basemap_source is provided
    if basemap_source is not None:
        ctx.add_basemap(ax, source=basemap_source)
        ax.set_axis_off()  # Optional: Remove axis for visual clarity

    # After plotting the GeoDataFrame
    prop_title, prop, prop_legend = _get_font_properties(font_size)
    ax.set_title(title, fontproperties=prop_title, color=font_color)  # Set the title with font properties

    if legend and not (plot_type == "point" and not variable_name):
        # Assuming variable_name represents continuous data. Adjust vmin and vmax accordingly.
        if variable_name:
            vmin, vmax = gdf[variable_name].min(), gdf[variable_name].max()
        else:
            vmin, vmax = gdf["count"].min(), gdf["count"].max()
        _add_colorbar(
            fig,
            ax,
            vmin,
            vmax,
            cmap,
            legend_title,
            prop,
            prop_legend,
            font_color,
            orientation="horizontal",
            dark_mode=dark_mode,
        )

    if path_output:
        plt.savefig(path_output, bbox_inches="tight", dpi=dpi)
    return fig, ax



# Copyright (c) Meta Platforms, Inc. and affiliates.
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.

from typing import Any, Tuple

from torch.utils.data import Dataset


class DatasetWithEnumeratedTargets(Dataset):
    """Dataset wrapper that provides enumerated targets.

    This class extends the PyTorch Dataset to include the index of the sample
    as part of the target, allowing for easier tracking of samples.

    Args:
        dataset (Dataset): The dataset to wrap.

    """

    def __init__(self, dataset: Dataset) -> None:
        """Initializes the DatasetWithEnumeratedTargets.

        Args:
            dataset (Dataset): The dataset to wrap.
        """
        self._dataset = dataset

    def get_image_data(self, index: int) -> bytes:
        """Retrieves the image data for a given index.

        Args:
            index (int): The index of the image to retrieve.

        Returns:
            bytes: The raw image data.
        """
        return self._dataset.get_image_data(index)

    def get_target(self, index: int) -> Tuple[Any, int]:
        """Retrieves the target data for a given index, including the index.

        Args:
            index (int): The index of the target to retrieve.

        Returns:
            Tuple[Any, int]: A tuple containing the index and the target data.
        """
        target = self._dataset.get_target(index)
        return (index, target)

    def __getitem__(self, index: int) -> Tuple[Any, Tuple[Any, int]]:
        """Fetches the image and target data for a given index.

        Args:
            index (int): The index of the data to retrieve.

        Returns:
            Tuple[Any, Tuple[Any, int]]: A tuple containing the image and a tuple
            of the index and target data.
        """
        image, target = self._dataset[index]
        target = index if target is None else target
        return image, (index, target)

    def __len__(self) -> int:
        """Returns the length of the dataset.

        Returns:
            int: The number of samples in the dataset.
        """
        return len(self._dataset)



# Copyright (c) Meta Platforms, Inc. and affiliates.
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.

from .dino_clstoken_loss import DINOLoss
from .ibot_patch_loss import iBOTPatchLoss
from .koleo_loss import KoLeoLoss

__all__ = ["DINOLoss", "iBOTPatchLoss", "KoLeoLoss"]



# Copyright (c) Facebook, Inc. and its affiliates. (http://www.facebook.com)
# -*- coding: utf-8 -*-
"""mapillary.models.api.vector_tiles
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

This module contains the adapter design for the Vector Tiles API of Mapillary API v4, Vector tiles
provide an easy way to visualize vast amounts of data. Vector tiles support filtering and querying
rendered features. Mapillary vector tiles follow the Mapbox tile specification.

For more information, please check out https://www.mapillary.com/developer/api-documentation/.

- Copyright: (c) 2021 Facebook
- License: MIT LICENSE
"""

import glob
import json
import os
from concurrent.futures import ThreadPoolExecutor, as_completed

import mercantile
from tqdm import tqdm

# Package imports
from vt2geojson.tools import vt_bytes_to_geojson

# Local imports
# # Config
from zensvi.download.mapillary.config.api.vector_tiles import VectorTiles

# # Client import
from zensvi.download.mapillary.models.client import Client

# # Exception handling
from zensvi.download.mapillary.models.exceptions import InvalidOptionError

# # Models
from zensvi.download.mapillary.models.geojson import GeoJSON


class VectorTilesAdapter(object):
    """Adapter model for dealing with the VectorTiles API, through the DRY principle.
    The VectorTilesAdapter class can be instantiated in the controller modules,
    providing an abstraction layer that uses the Client class, endpoints provided by the
    API v4 under `/config/api/vector_tiles.py`.

    It performs parsing, handling of layers, properties, and fields to make it easier to
    write higher level logic for extracing information, and lets developers to focus only
    on writing the high level business logic without having to repeat the process of parsing
    and using libraries such as `mercantile`, 'vt2geojson' and others, caring only about
    inputs/outputs

    Usage::

        >>> import mapillary
        >>> from mapillary.models.api.vector_tiles import VectorTilesAdapter
        >>> latitude, longitude = 30, 31
        >>> VectorTilesAdapter().fetch_layer(layer="image", zoom=14, longitude=longitude,
        ...     latitude=latitude,
        ... )
        >>> VectorTilesAdapter().fetch_layer(layer="sequence", zoom=10, longitude=longitude,
        ...     latitude=latitude,
        ... )
        >>> VectorTilesAdapter().fetch_layer(layer="overview", zoom=3, longitude=longitude,
        ...     latitude=latitude,
        ... )
    """

    # FOR DEVELOPERS
    # Future changes here will most likely work on the following aspects,
    # 1. Adjusting zoom levels
    # 1.1. Most likely changes in __zoom_range_check
    # 2. The zoom levels themselves
    # 2.1. Most likely changes in __init__
    # 3. Preprocessing steps depending on the layer targeted
    # 3.1. Most likely changes in __preprocess_layer, __preprocess_computed_layer,
    # __preprocess_features

    def __init__(self) -> None:
        """Initializing VectorTilesAdapter constructor."""
        # Initialize client object
        self.client = Client()

        # * See "FOR DEVELOPERS (2, 2.1)"

        # Setting the max zoom value
        self.__min_zoom = 0

        # Setting the min zoom value
        self.__max_zoom = 14

    def fetch_layer(self, layer: str, longitude: float, latitude: float, zoom: int = 14) -> dict:
        """Fetches an image tile layer depending on the coordinates, and the layer
        selected along with the zoom level.

        Args:
            layer (str): Either 'overview', 'sequence', 'image'
            longitude (float): The longitude of the coordinates
            latitude (float): The latitude of the coordinates
            zoom (int): The zoom level, [0, 14], inclusive

        Returns:
            dict: A GeoJSON for that specific layer and the specified
            zoom level
        """
        # Checking if the correct parameters are passed
        VectorTilesAdapter.__check_parameters(longitude=longitude, latitude=latitude)

        # Check for the correct zoom values against the layer specified
        self.__zoom_range_check(layer=layer, zoom=zoom)

        # Return the results of the layer after preprocessing steps
        return self.__preprocess_layer(
            # The layer to retrieve from
            layer=layer,
            # Turn coordinates into a tile
            tile=mercantile.tile(lng=longitude, lat=latitude, zoom=zoom),
            # The zoom level
            zoom=zoom,
        )

    def fetch_computed_layer(self, layer: str, zoom: int, longitude: float, latitude: float):
        """Same as `fetch_layer`, but gets in return computed tiles only. Depends on the
        layer, zoom level, longitude and the latitude specifications.

        Args:
            layer (str): Either 'overview', 'sequence', 'image'
            zoom (int): The zoom level, [0, 14], inclusive
            longitude (float): The longitude of the coordinates
            latitude (float): The latitude of the coordinates

        Returns:
            dict: A GeoJSON for that specific layer and the specified
            zoom level
        """
        # Checking if the correct parameters are passed
        VectorTilesAdapter.__check_parameters(longitude=longitude, latitude=latitude)

        # Check for the correct zoom values against the layer specified
        self.__zoom_range_check(layer=layer, zoom=zoom)

        # Return the results of the layer after preprocessing steps
        return self.__preprocess_computed_layer(
            # The layer to retrieve from
            layer=layer,
            # Turn coordinates into a tile
            tile=mercantile.tile(lng=longitude, lat=latitude, zoom=zoom),
            # The zoom level
            zoom=zoom,
        )

    def fetch_features(self, feature_type: str, zoom: int, longitude: float, latitude: float):
        """Fetches specified features from the coordinates with the appropriate zoom
        level.

        Args:
            feature_type (str): Either `point`, or `tiles`
            zoom (int): The zoom level
            longitude (float): The longitude of the coordinates
            latitude (float): The latitude of the coordinates

        Returns:
            dict: A GeoJSON for that specific layer and the specified
            zoom level
        """
        # Checking if the correct parameters are passed
        VectorTilesAdapter.__check_parameters(longitude=longitude, latitude=latitude)

        # Check for the correct zoom values against the layer specified
        self.__zoom_range_check(layer="map", zoom=zoom)

        # Return the results of the layer after preprocessing steps
        return self.__preprocess_features(
            # The feature type to retrieve for
            feature_type=feature_type,
            # Turn coordinates into a tile
            tile=mercantile.tile(lng=longitude, lat=latitude, zoom=zoom),
            # The zoom level
            zoom=zoom,
        )

    def fetch_layers(
        self,
        coordinates: "list[list]",
        layer: str = "image",
        zoom: int = 14,
        is_computed: bool = False,
        **kwargs,
    ) -> GeoJSON:
        """Fetches multiple vector tiles based on a list of multiple coordinates in a
        listed format.

        Args:
            coordinates ("list[list]"): A list of lists of coordinates
                to get the vector tiles for
            layer (str): Either "overview", "sequence", "image",
                "traffic_sign", or "map_feature", defaults to "image"
            zoom (int): the zoom level [0, 14], inclusive. Defaults to
                14
            is_computed (bool): Will to be fetched layers be computed?
                Defaults to False

        Returns:
            dict: A geojson with merged features from all unique vector
            tiles
        """
        # Check for the correct zoom values against the layer specified
        self.__zoom_range_check(layer=layer, zoom=zoom)

        # The output resultant geojson
        geojson: GeoJSON = GeoJSON(geojson={"type": "FeatureCollection", "features": []})

        # A list of tiles that are either confined within or intersect with the bbox
        tiles = list(
            mercantile.tiles(
                west=coordinates[0],
                south=coordinates[1],
                east=coordinates[2],
                north=coordinates[3],
                zooms=zoom,
            )
        )

        # Helper function to process each tile
        def process_tile(tile):
            if is_computed:
                result = self.__preprocess_computed_layer(layer=layer, tile=tile, zoom=zoom)["features"]
            else:
                result = self.__preprocess_layer(layer=layer, tile=tile, zoom=zoom)["features"]
            return result

        # Create a directory for checkpoints and load existing results
        dir_cache_tiles = None
        processed_tiles = set()

        # Function to process each file
        def process_file(file_path):
            try:
                with open(file_path, "r") as file:
                    tile_data = json.load(file)
                if tile_data["features"]:
                    return (
                        tile_data["features"],
                        file_path,
                        False,
                    )  # last argument is whether there was an error
                else:
                    return None, file_path, False
            except Exception as e:
                print(f"Error loading file {file_path}: {e}")
            return None, file_path, True

        if kwargs["dir_cache"]:
            dir_cache_tiles = os.path.join(kwargs["dir_cache"], "tiles_results")
            os.makedirs(dir_cache_tiles, exist_ok=True)

            existing_files = glob.glob(f"{dir_cache_tiles}/*.geojson")

            # Use ThreadPoolExecutor to process files in parallel
            with ThreadPoolExecutor(max_workers=kwargs["max_workers"]) as executor:
                future_to_file = {executor.submit(process_file, file_path): file_path for file_path in existing_files}
                for future in tqdm(
                    as_completed(future_to_file),
                    total=len(existing_files),
                    desc="Loading cache files",
                ):
                    result, file_path, error = future.result()
                    if result:
                        geojson.append_features(result)
                    if not error:
                        filename = os.path.basename(file_path)
                        x, y, z = map(int, filename.replace(".geojson", "").split("_"))
                        processed_tiles.add((x, y, z))

            # Filter out tiles that have already been processed
            tiles = [tile for tile in tiles if (tile.x, tile.y, tile.z) not in processed_tiles]

        if len(tiles) == 0:
            print("All tiles have already been processed.")
            return geojson

        print(f'[Vector Tiles API] Fetching {len(tiles)} {"tiles" if len(tiles) > 1 else "tile"}' " for images ...")
        # Define the maximum number of retries for processing a tile
        max_retries = 3
        # Process each tile individually: set max_workers to 10 times the number of CPU cores
        with ThreadPoolExecutor(max_workers=kwargs["max_workers"]) as executor:
            future_to_tile = {executor.submit(process_tile, tile): tile for tile in tiles}
            for future in tqdm(as_completed(future_to_tile), total=len(tiles), desc="Processing tiles"):
                retries = 0
                while retries <= max_retries:
                    try:
                        tile = future_to_tile[future]
                        result = future.result()
                        # append if len(result) > 0
                        if result:
                            geojson.append_features(result)
                        if kwargs["dir_cache"]:
                            # Save result for each tile
                            result_path = f"{dir_cache_tiles}/{tile.x}_{tile.y}_{tile.z}.geojson"
                            with open(result_path, "w") as file:
                                json.dump(
                                    {"type": "FeatureCollection", "features": result},
                                    file,
                                )
                        break  # Break the loop if processing is successful
                    except Exception as e:
                        print(f"Error processing tile {tile.x}_{tile.y}_{tile.z} on attempt {retries + 1}: {e}")
                        retries += 1
                        if retries > max_retries:
                            print(f"Failed to process tile {tile.x}_{tile.y}_{tile.z} after {max_retries} attempts.")
                            if kwargs["logger"]:
                                kwargs["logger"].log_failed_tiles(f"{tile.x}_{tile.y}_{tile.z}")

        return geojson

    def fetch_map_features(
        self,
        coordinates: "list[list]",
        feature_type: str,
        zoom: int = 14,
    ) -> GeoJSON:
        """Fetches map features based on a list Polygon object.

        Args:
            coordinates ("list[list]"): A list of lists of coordinates
                to get the map features for
            feature_type (str): Either "point", "traffic_signs",
                defaults to "point"
            zoom (int): the zoom level [0, 14], inclusive. Defaults to
                14

        Returns:
            dict: A geojson with merged features from all unique vector
            tiles
        """
        # Check for the correct zoom values against the layer specified
        self.__zoom_range_check(layer="map_feature", zoom=zoom)

        # The output resultant geojson
        geojson: GeoJSON = GeoJSON(geojson={"type": "FeatureCollection", "features": []})

        # A list of tiles that are either confined within or intersect with the bbox
        tiles = list(
            mercantile.tiles(
                west=coordinates[0],
                south=coordinates[1],
                east=coordinates[2],
                north=coordinates[3],
                zooms=zoom,
            )
        )

        print(
            f'[Vector Tiles API] Fetching {len(tiles)} {"tiles" if len(tiles) > 1 else "tile"}' " for map features ..."
        )

        for tile in tiles:
            geojson.append_features(
                self.__preprocess_features(
                    # The layer to retrieve from
                    feature_type=feature_type,
                    # Turn coordinates into a tile
                    tile=tile,
                    # The zoom level
                    zoom=zoom,
                )["features"]
            )

        return geojson

    @staticmethod
    def __check_parameters(
        longitude: float,
        latitude: float,
    ):
        """Range checking for the parameters of longitude, latitude, layer, zoom.

        Args:
            longitude (float): The longitude of the coordinates
            latitude (float): The latitude of the coordinates

        Returns:
            dict: A GeoJSON for that specific layer and the specified
            zoom level
        """
        # Lng, Lat ranges, https://docs.mapbox.com/help/glossary/lat-lon/

        # If lng not in the range [-180, 180], inclusive
        if longitude <= -180 or longitude >= 180:
            # Raise exception
            raise InvalidOptionError(param="longitude", value=longitude, options=[-180, 180])

        # If lat not in the range [-90, 90], inclusive
        if latitude <= -90 or latitude >= 90:
            # Raise exception
            raise InvalidOptionError(param="latitude", value=latitude, options=[-180, 180])

    def __zoom_range_check(self, layer: str, zoom: int):
        """Checks for the correct zoom values for te specified layer.

        Args:
            layer (str): Either 'overview', 'sequence', 'image', or
                'map'
            zoom (int): The zoom levels,

        Raises:
            InvalidOptionError: Invalid option passed

        Returns:
            dict: A GeoJSON for the return object
        """
        # If zoom is not in the valid range of values
        if zoom < self.__min_zoom or zoom > self.__max_zoom:
            # Raise an exception for the invalid values passed
            raise InvalidOptionError(
                # Parameters accordingly
                param="zoom",
                value=zoom,
                options=[_ for _ in range(self.__min_zoom, self.__max_zoom + 1)],
            )

        # If layer was specified to be 'overview'
        if layer == "overview":

            # If zoom is not in the given range of values
            if zoom not in [_ for _ in range(self.__min_zoom, 5 + 1)]:
                # Raise an exception for the invalid zoom value
                raise InvalidOptionError(
                    # Parameters accordingly
                    param="zoom",
                    value=zoom,
                    options=[_ for _ in range(self.__min_zoom, 5 + 1)],
                )

        # If layer was specified to be 'sequence'
        elif layer == "sequence":

            # IF zoom not in the given range of values
            if zoom not in [_ for _ in range(6, 14 + 1)]:
                # Raise an exception for the invalid zoom value
                raise InvalidOptionError(
                    # Parameters accordingly
                    param="zoom",
                    value=zoom,
                    options=[_ for _ in range(6, 14 + 1)],
                )

        # If layer was specified to be 'image' or 'map'
        elif layer == "image" or layer == "map_feature" or layer == "traffic_sign":

            # If zoom was not 14
            if zoom != 14:
                # Raise an exception for the invalid zoom value
                raise InvalidOptionError(param="zoom", value=zoom, options=[14])

        # Else, an invalid layer string was passed
        else:

            # Raise a InvalidOptionError
            raise InvalidOptionError(
                # Parameters accordingly
                param="layer",
                value=layer,
                options=["overview", "sequence", "image"],
            )

    def __preprocess_layer(self, layer: str, tile: mercantile.Tile, zoom: int):
        """Preprocessing un-computed layers.

        Args:
            layer (str): Either 'overview', 'sequence', 'image'
            tile (mercantile.Tile): The specified tile
            zoom (int): The zoom level

        Returns:
            dict: A GeoJSON
        """
        # * See "FOR DEVELOPERS (3, 3.1)"

        # Extract the url depending upon the layer
        url = ""

        # For overview
        if layer == "overview":
            url = VectorTiles.get_overview_layer(x=tile[0], y=tile[1], z=zoom)

        # For sequence
        elif layer == "sequence":
            url = VectorTiles.get_sequence_layer(x=tile[0], y=tile[1], z=zoom)

        # For image
        elif layer == "image":
            url = VectorTiles.get_image_layer(x=tile[0], y=tile[1], z=zoom)

        # No 'else' for InvalidOptionError, as checking done previously in __zoom_range_check

        # Convert bytes to GeoJSON
        return vt_bytes_to_geojson(
            # Parameters, appropriately
            b_content=self.client.get(url).content,
            x=tile.x,
            y=tile.y,
            z=tile.z,
            layer=layer,
        )

    def __preprocess_computed_layer(self, layer: str, tile: mercantile.Tile, zoom: int):
        """Preprocessing computed layers.

        Args:
            layer (str): Either 'overview', 'sequence', 'image'
            tile (mercantile.Tile): The specified tile
            zoom (int): The zoom level

        Returns:
            dict: A GeoJSON
        """
        # * See "FOR DEVELOPERS (3, 3.1)"

        # Extracting url from specified VectorTiles endpoint
        url = ""

        # For overview
        if layer == "overview":
            url = VectorTiles.get_computed_overview_layer(x=tile[0], y=tile[1], z=zoom)

        # For sequence
        elif layer == "sequence":
            url = VectorTiles.get_computed_sequence_layer(x=tile[0], y=tile[1], z=zoom)

        # For image
        elif layer == "image":
            url = VectorTiles.get_computed_image_layer(x=tile[0], y=tile[1], z=zoom)

        # No 'else' for InvalidOptionError, as checking done previously in __zoom_range_check

        # Convert bytes to geojson
        return vt_bytes_to_geojson(
            # Parameters appropriately
            b_content=self.client.get(url).content,
            x=tile.x,
            y=tile.y,
            z=tile.z,
            layer=layer,
        )

    def __preprocess_features(self, feature_type: str, tile: mercantile.Tile, zoom: int) -> dict:
        """Preprocess features.

        Args:
            feature_type (str): Either 'point', 'traffic_signs'
            tile (mercantile.Tile): The specified tile
            zoom (int): The zoom level

        Returns:
            dict: A GeoJSON
        """
        # * See "FOR DEVELOPERS (3, 3.1)"

        # Extracting url from specified VectorTiles endpoint

        # For 'point'
        if feature_type == "point":

            url = VectorTiles.get_map_feature_point(x=tile[0], y=tile[1], z=zoom)

        # For 'traffic_signs'
        elif feature_type == "traffic_signs":
            url = VectorTiles.get_map_feature_traffic_sign(x=tile[0], y=tile[1], z=zoom)

        # If both are not present
        else:
            # Raise an exception for the invalid values passed
            raise InvalidOptionError(
                # Parameters accordingly
                param="feature_type",
                value=feature_type,
                options=["point", "traffic_sign"],
            )

        # Convert bytes to GeoJSON, and return
        return vt_bytes_to_geojson(
            # Parameters appropriately
            b_content=self.client.get(url).content,
            x=tile.x,
            y=tile.y,
            z=tile.z,
            layer=None,
        )



# Copyright (c) Meta Platforms, Inc. and affiliates.
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.

import torch
import torch.nn as nn
from torch.nn.init import trunc_normal_
from torch.nn.utils import weight_norm


class DINOHead(nn.Module):
    """DINO Head for the DINO model.

    This class implements the DINO head which consists of a multi-layer perceptron (MLP)
    followed by a normalization layer and a final linear layer.

    Args:
        in_dim (int): The input dimension of the MLP.
        out_dim (int): The output dimension of the final layer.
        use_bn (bool, optional): Whether to use batch normalization in the MLP. Defaults to False.
        nlayers (int, optional): The number of layers in the MLP. Defaults to 3.
        hidden_dim (int, optional): The hidden dimension of the MLP. Defaults to 2048.
        bottleneck_dim (int, optional): The bottleneck dimension of the MLP. Defaults to 256.
        mlp_bias (bool, optional): Whether to use bias in the MLP layers. Defaults to True.
    """

    def __init__(
        self,
        in_dim,
        out_dim,
        use_bn=False,
        nlayers=3,
        hidden_dim=2048,
        bottleneck_dim=256,
        mlp_bias=True,
    ):
        super().__init__()
        nlayers = max(nlayers, 1)
        self.mlp = _build_mlp(
            nlayers,
            in_dim,
            bottleneck_dim,
            hidden_dim=hidden_dim,
            use_bn=use_bn,
            bias=mlp_bias,
        )
        self.apply(self._init_weights)
        self.last_layer = weight_norm(nn.Linear(bottleneck_dim, out_dim, bias=False))
        self.last_layer.weight_g.data.fill_(1)

    def _init_weights(self, m):
        """Initializes the weights of the model.

        This method applies weight initialization to the linear layers.

        Args:
            m (nn.Module): The module to initialize.
        """
        if isinstance(m, nn.Linear):
            trunc_normal_(m.weight, std=0.02)
            if m.bias is not None:
                nn.init.constant_(m.bias, 0)

    def forward(self, x):
        """Forward pass for the DINO head.

        Args:
            x (torch.Tensor): The input tensor.

        Returns:
            torch.Tensor: The output tensor after applying the MLP and final linear layer.
        """
        x = self.mlp(x)
        eps = 1e-6 if x.dtype == torch.float16 else 1e-12
        x = nn.functional.normalize(x, dim=-1, p=2, eps=eps)
        x = self.last_layer(x)
        return x


def _build_mlp(nlayers, in_dim, bottleneck_dim, hidden_dim=None, use_bn=False, bias=True):
    """Builds a multi-layer perceptron (MLP).

    This function constructs an MLP with the specified number of layers, input dimension,
    bottleneck dimension, and optional batch normalization.

    Args:
        nlayers (int): The number of layers in the MLP.
        in_dim (int): The input dimension of the MLP.
        bottleneck_dim (int): The bottleneck dimension of the MLP.
        hidden_dim (int, optional): The hidden dimension of the MLP. Defaults to None.
        use_bn (bool, optional): Whether to use batch normalization in the MLP. Defaults to False.
        bias (bool, optional): Whether to use bias in the MLP layers. Defaults to True.

    Returns:
        nn.Sequential: A sequential container of the MLP layers.
    """
    if nlayers == 1:
        return nn.Linear(in_dim, bottleneck_dim, bias=bias)
    else:
        layers = [nn.Linear(in_dim, hidden_dim, bias=bias)]
        if use_bn:
            layers.append(nn.BatchNorm1d(hidden_dim))
        layers.append(nn.GELU())
        for _ in range(nlayers - 2):
            layers.append(nn.Linear(hidden_dim, hidden_dim, bias=bias))
            if use_bn:
                layers.append(nn.BatchNorm1d(hidden_dim))
            layers.append(nn.GELU())
        layers.append(nn.Linear(hidden_dim, bottleneck_dim, bias=bias))
        return nn.Sequential(*layers)



# Copyright (c) Meta Platforms, Inc. and affiliates.
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.

import argparse
from typing import Any, List, Optional, Tuple

import dinov2.utils.utils as dinov2_utils
import torch
import torch.backends.cudnn as cudnn
from dinov2.models import build_model_from_cfg
from dinov2.utils.config import setup


def get_args_parser(
    description: Optional[str] = None,
    parents: Optional[List[argparse.ArgumentParser]] = None,
    add_help: bool = True,
) -> argparse.ArgumentParser:
    """Creates an argument parser for the evaluation setup.

    Args:
        description (Optional[str]): Description of the parser.
        parents (Optional[List[argparse.ArgumentParser]]): Parent parsers to inherit arguments from.
        add_help (bool): Whether to add a help option to the parser.

    Returns:
        argparse.ArgumentParser: The configured argument parser.
    """
    parser = argparse.ArgumentParser(
        description=description,
        parents=parents or [],
        add_help=add_help,
    )
    parser.add_argument(
        "--config-file",
        type=str,
        help="Model configuration file",
    )
    parser.add_argument(
        "--pretrained-weights",
        type=str,
        help="Pretrained model weights",
    )
    parser.add_argument(
        "--output-dir",
        default="",
        type=str,
        help="Output directory to write results and logs",
    )
    parser.add_argument(
        "--opts",
        help="Extra configuration options",
        default=[],
        nargs="+",
    )
    return parser


def get_autocast_dtype(config) -> torch.dtype:
    """Determines the appropriate autocast data type based on the configuration.

    Args:
        config: The configuration object containing precision settings.

    Returns:
        torch.dtype: The data type to use for autocasting.
    """
    teacher_dtype_str = config.compute_precision.teacher.backbone.mixed_precision.param_dtype
    if teacher_dtype_str == "fp16":
        return torch.half
    elif teacher_dtype_str == "bf16":
        return torch.bfloat16
    else:
        return torch.float


def build_model_for_eval(config, pretrained_weights) -> torch.nn.Module:
    """Builds the model for evaluation using the provided configuration and weights.

    Args:
        config: The configuration object for the model.
        pretrained_weights: Path to the pretrained model weights.

    Returns:
        torch.nn.Module: The constructed model ready for evaluation.
    """
    model, _ = build_model_from_cfg(config, only_teacher=True)
    dinov2_utils.load_pretrained_weights(model, pretrained_weights, "teacher")
    model.eval()
    model.cuda()
    return model


def setup_and_build_model(args) -> Tuple[Any, torch.dtype]:
    """Sets up the model and returns it along with the autocast data type.

    Args:
        args: The command line arguments containing configuration and weights.

    Returns:
        Tuple[Any, torch.dtype]: The model and the autocast data type.
    """
    cudnn.benchmark = True
    config = setup(args)
    model = build_model_for_eval(config, args.pretrained_weights)
    autocast_dtype = get_autocast_dtype(config)
    return model, autocast_dtype



# Copyright (c) Meta Platforms, Inc. and affiliates.
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.

import re
from pathlib import Path
from typing import List, Tuple

from setuptools import find_packages, setup

NAME = "dinov2"
DESCRIPTION = "PyTorch code and models for the DINOv2 self-supervised learning method."

URL = "https://github.com/facebookresearch/dinov2"
AUTHOR = "FAIR"
REQUIRES_PYTHON = ">=3.9.0"
HERE = Path(__file__).parent


try:
    with open(HERE / "README.md", encoding="utf-8") as f:
        long_description = "\n" + f.read()
except FileNotFoundError:
    long_description = DESCRIPTION


def get_requirements(
    path: str = HERE / "requirements.txt",
) -> Tuple[List[str], List[str]]:
    """Retrieve the package requirements from a specified file.

    Args:
        path (str): The path to the requirements file.
                    Defaults to HERE / "requirements.txt".

    Returns:
        Tuple[List[str], List[str]]: A tuple containing a list of
        requirements and a list of extra index URLs.
    """
    requirements = []
    extra_indices = []
    with open(path) as f:
        for line in f.readlines():
            line = line.rstrip("\r\n")
            if line.startswith("--extra-index-url "):
                extra_indices.append(line[18:])
                continue
            requirements.append(line)
    return requirements, extra_indices


def get_package_version() -> str:
    """Get the package version from the __init__.py file.

    Returns:
        str: The package version.

    Raises:
        RuntimeError: If the package version cannot be determined.
    """
    with open(HERE / "dinov2/__init__.py") as f:
        result = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]", f.read(), re.M)
        if result:
            return result.group(1)
    raise RuntimeError("Can't get package version")


requirements, extra_indices = get_requirements()
version = get_package_version()
dev_requirements, _ = get_requirements(HERE / "requirements-dev.txt")


setup(
    name=NAME,
    version=version,
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type="text/markdown",
    author=AUTHOR,
    python_requires=REQUIRES_PYTHON,
    url=URL,
    packages=find_packages(),
    package_data={
        "": ["*.yaml"],
    },
    install_requires=requirements,
    dependency_links=extra_indices,
    extras_require={
        "dev": dev_requirements,
    },
    install_package_data=True,
    license="CC-BY-NC",
    license_files=("LICENSE",),
    classifiers=[
        # Trove classifiers: https://github.com/pypa/trove-classifiers/blob/main/src/trove_classifiers/__init__.py
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "License :: Other/Proprietary License",
        "Programming Language :: Python :: 3.9",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
)
